"""
Topology transformations
========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains algorithms for initializing or transforming
topologies.
"""

from typing import Any, Union
import warnings

import MDAnalysis as mda
from MDAnalysis.lib.distances import minimize_vectors
from MDAnalysis.lib.mdamath import make_whole
import numpy as np

from .. import FOUND_OPENMM
from .molecule import center_of_mass
from .utility import get_closest_factors, replicate, find_connected_nodes
from .unit import strip_unit

if FOUND_OPENMM:
    from openmm import app, unit

def create_atoms(
        dims: Union[np.ndarray[float], "unit.Quantity", "app.Topology"],
        N: int = None, N_p: int = 1, *, lattice: str = None,
        length: Union[float, "unit.Quantity"] = 0.34,
        flexible: bool = False, bonds: bool = False, angles: bool = False,
        dihedrals: bool = False, randomize: bool = False,
        length_unit: "unit.Unit" = None, wrap: bool = False) -> Any:

    r"""
    Generates initial particle positions for coarse-grained simulations.

    Parameters
    ----------
    dims : `numpy.ndarray`, `openmm.unit.Quantity`, or `openmm.app.Topology`
        System dimensions, provided as an array or obtained from an
        OpenMM topology.

        **Reference unit**: :math:`\mathrm{nm}`.

    N : `int`, optional
        Total number of particles (monomers). Must be provided for
        random melts or polymers.

    N_p : `int`, default: :code:`1`
        Number of particles (monomers) :math:`N_p` in each
        segment (polymer chain).

        **Valid values**: :math:`1\leq N_\mathrm{p}\leq N`, with
        :math:`N` divisible by :math:`N_\mathrm{p}`.

    lattice : `str`, keyword-only, optional
        Lattice type, with the relevant length scale specified in
        `length`. If `lattice` is not specified, particle positions will
        be assigned randomly.

        .. tip::

           To build walls with the correct periodicity, set the
           :math:`z`-dimension to :code:`0` in `dims` and
           :code:`flexible=True`. This function will then return
           the wall particle positions and the :math:`x`- and
           :math:`y`-dimensions closest to those specified in `dims`
           that also satisfy the lattice periodicity.

           Keep in mind that walls should only be built in the
           :math:`z`-direction.

        .. container::

           **Valid values**:

           * :code:`"fcc"`: Face-centered cubic (FCC) lattice,
             determined by the particle size :math:`a`.
           * :code:`"hcp"`: Hexagonal close-packed (HCP) lattice,
             determined by the particle size :math:`a`.
           * :code:`"cubic"`: Cubic crystal system, determined by the
             particle size :math:`a`.
           * :code:`"honeycomb"`: Honeycomb lattice (e.g., graphene),
             determined by the bond length :math:`b`.

    length : `float` or `openmm.unit.Quantity`, default: :code:`0.34`
        For random polymer positions, `length` is the bond length used in
        the random walk. For lattice systems, `length` is either the
        particle size or the bond length, depending on the lattice type.
        Has no effect if :code:`N_p=1` or :code:`lattice=None`.

        **Reference unit**: :math:`\mathrm{nm}`.

    flexible : `bool`, default: :code:`False`
        Specifies whether the dimensions provided in `dims` can be
        changed to satisfy the lattice periodicity, if applicable.
        For example, if the provided :math:`z`-dimension can hold a
        non-integer 19.7 replicas of a lattice, then it is updated
        to reflect the width of 20 replicas. To ignore a direction (and
        make that dimension constant), such as when creating walls, set
        that dimension to :code:`0` in `dims`.

    bonds : `bool`, default: :code:`False`
        Determines whether bond information is returned for polymeric
        systems. Has no effect if :code:`N_p=1`.

    angles : `bool`, default: :code:`False`
        Determines whether angle information is returned for polymeric
        systems. Has no effect if :code:`N_p=1`.

    dihedrals : `bool`, default: :code:`False`
        Determines whether dihedral information is returned for polymeric
        systems. Has no effect if :code:`N_p=1`.

    randomize : `bool`, default: :code:`False`
        Determines whether the order of the replicated polymer positions
        are randomized. Has no effect if :code:`N_p=1`.

    length_unit : `openmm.unit.Unit`, optional
        Length unit. If not specified, it is determined automatically
        from `dims` or `length`.

    wrap : `bool`, default: :code:`False`
        Determines whether particles outside the simulation box are
        wrapped back into the main unit cell.

    Returns
    -------
    positions : `numpy.ndarray` or `openmm.unit.Quantity`
        Generated particle positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    bonds : `numpy.ndarray`
        Pairs of all bonded particle indices. Only returned if
        :code:`bonds=True`.

        **Shape**: :math:`(N_\mathrm{bonds},\,2)`.

    angles : `numpy.ndarray`
        Triples of all particle indices that form an angle. Only
        returned if :code:`angles=True`.

        **Shape**: :math:`(N_\mathrm{angles},\,3)`.

    dihedrals : `numpy.ndarray`
        Quadruples of all particle indices that form a dihedral. Only
        returned if :code:`dihedrals=True`.

        **Shape**: :math:`(N_\mathrm{dihedrals},\,4)`.

    dimensions : `numpy.ndarray` or `openmm.unit.Quantity`
        Dimensions for lattice systems. Only returned if `lattice` is
        not :code:`None`.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.
    """

    # Get raw numerical dimensions and length
    if isinstance(dims, app.Topology):
        dims = dims.getUnitCellDimensions()
    dims, length_unit = strip_unit(dims, length_unit)
    length, length_unit = strip_unit(length, length_unit)
    length_unit = length_unit or 1

    if lattice is None:

        # Ensure the user-specified values are valid
        if N is None:
            raise ValueError("The number of particles N must be specified.")
        if not isinstance(N, (int, np.integer)):
            raise ValueError("The number of particles N must be an integer.")
        if not (1 <= N_p <= N and isinstance(N_p, (int, np.integer))):
            emsg = ("The number of particles N_p in each segment must "
                    "be an integer between 1 and N.")
            raise ValueError(emsg)
        if N_p > 1 and N % N_p != 0:
            emsg = (f"{N=} particles cannot be evenly divided into segments "
                    f"with {N_p=} particles.")
            raise ValueError(emsg)

        # Generate particle positions for a random melt
        if N_p == 1:
            return np.random.rand(N, 3) * dims * length_unit
        else:
            topo = []

            # Determine unit cell information for each segment
            segments = N // N_p
            n_cells = get_closest_factors(segments, 3)
            cell_dims = dims / n_cells

            # Randomly generate a segment within the unit cell
            cell_pos = np.zeros((N_p, 3))
            cell_pos[0] = cell_dims / 4
            rng = np.random.default_rng()
            for i in range(1, N_p):
                vec = rng.random(3) * 2 - 1
                cell_pos[i] = cell_pos[i - 1] + length * vec / np.linalg.norm(vec)

            # Replicate unit cell in x-, y-, and z-directions (and
            # randomize, if desired)
            pos = replicate(cell_dims, cell_pos, n_cells)
            if randomize:
                pos = np.vstack(rng.permutation(pos.reshape((segments, -1, 3))))

            # Wrap particles past the system boundaries
            if wrap:
                for i in range(3):
                    pos[pos[:, i] < 0, i] += dims[i]
                    pos[pos[:, i] > dims[i], i] -= dims[i]

            topo.append(pos * length_unit)

            # Determine all bonds
            if bonds:
                topo.append(np.array([(i * N_p + j, i * N_p + j + 1)
                                      for i in range(segments)
                                      for j in range(N_p - 1)]))

            # Determine all angles
            if angles:
                topo.append(np.array([np.arange(i * N_p + j, i * N_p + j + 3)
                                      for i in range(segments)
                                      for j in range(N_p - 2)]))

            # Determine all dihedrals
            if dihedrals:
                topo.append(np.array([np.arange(i * N_p + j, i * N_p + j + 4)
                                      for i in range(segments)
                                      for j in range(N_p - 3)]))

            return topo[0] if len(topo) == 1 else tuple(topo)
    else:
        around = np.around if flexible else np.floor

        # Set unit cell information
        if lattice == "cubic":
            _dims = np.array(dims)
            _dims[dims == 0] = 1
            n_cells = around(_dims / length).astype(int)
            cell_dims = length * np.array((1, 1, 1))
            x, y, z = (length * np.arange(n) for n in n_cells)
            pos = np.stack(np.meshgrid(x, y, z), axis=-1).reshape(-1, 3)
        else:
            if lattice == "fcc":
                cell_dims = length * np.array((1, np.sqrt(3), 3 * np.sqrt(6) / 3))
                cell_pos = length * np.array((
                    (0, 0, 0),
                    (0.5, np.sqrt(3) / 2, 0),
                    (0.5, np.sqrt(3) / 6, np.sqrt(6) / 3),
                    (0, 2 * np.sqrt(3) / 3, np.sqrt(6) / 3),
                    (0, np.sqrt(3) / 3, 2 * np.sqrt(6) / 3),
                    (0.5, 5 * np.sqrt(3) / 6, 2 * np.sqrt(6) / 3),
                ))
            elif lattice == "hcp":
                cell_dims = length * np.array((1, np.sqrt(3), 2 * np.sqrt(6) / 3))
                cell_pos = length * np.array((
                    (0, 0, 0),
                    (0.5, np.sqrt(3) / 2, 0),
                    (0.5, np.sqrt(3) / 6, np.sqrt(6) / 3),
                    (0, 2 * np.sqrt(3) / 3, np.sqrt(6) / 3)
                ))
            elif lattice == "honeycomb":
                cell_dims = length * np.array((np.sqrt(3), 3, np.inf))
                cell_pos = length * np.array((
                    (0, 0, 0),
                    (0, 1, 0),
                    (np.sqrt(3) / 2, 1.5, 0),
                    (np.sqrt(3) / 2, 2.5, 0)
                ))

            # Determine unit cell multiples
            n_cells = around(dims / cell_dims).astype(int)
            n_cells[n_cells == 0] = 1
            cell_dims[np.isinf(cell_dims)] = 0

            # Replicate unit cell in x-, y-, and z-directions
            pos = replicate(cell_dims, cell_pos, n_cells)

        # Remove particles outside of system boundaries
        if flexible:
            n_cells[dims == 0] = 0
            pos = pos[~np.any(pos[:, dims == 0] > 0, axis=1)]
        else:
            pos = pos[~np.any(pos > dims, axis=1)]

        return pos * length_unit, n_cells * cell_dims * length_unit

def unwrap(
        positions: np.ndarray[float], positions_old: np.ndarray[float],
        dimensions: np.ndarray[float], *, thresholds: float = None,
        images: np.ndarray[int] = None, in_place: bool = True
    ) -> Union[None, tuple[np.ndarray[float], np.ndarray[float],
                           np.ndarray[int]]]:

    r"""
    Globally unwraps particle positions.

    Parameters
    ----------
    positions : `numpy.ndarray`
        Particle positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    positions_old : `numpy.ndarray`
        Previous particle positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    dimensions : `numpy.ndarray`
        System dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    thresholds : `float`, keyword-only, optional
        Maximum distances in each direction a particle can move before
        it is considered to have crossed a boundary.

        **Reference unit**: :math:`\mathrm{nm}`.

    images : `numpy.ndarray`, keyword-only, optional
        Current image flags (or number of boundary crossings).

        **Shape**: :math:`(N,\,3)`.

    in_place : `bool`, keyword-only, default: :code:`False`
        Determines whether the input array is modified in-place.

    Returns
    -------
    positions : `numpy.ndarray`
        Unwrapped particle positions. Only returned if
        :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    positions_old : `numpy.ndarray`
        Updated previous particle positions. Only returned if
        :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    images : `numpy.ndarray`
        Updated image flags (or number of boundary crossings). Only
        returned if :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.
    """

    if thresholds is None:
        thresholds = np.min(dimensions) / 2
    if images is None:
        images = np.zeros_like(positions, dtype=int)

    dpos = positions - positions_old
    mask = np.abs(dpos) >= thresholds
    if in_place:
        images[mask] -= np.sign(dpos[mask]).astype(int)
        positions_old[:] = positions[:]
        positions += images * dimensions
    else:
        positions = positions.copy()
        positions_old = positions.copy()
        images = images.copy()
        images[mask] -= np.sign(dpos[mask]).astype(int)
        positions += images * dimensions
        return positions, positions_old, images

def unwrap_edge(
        *, group: mda.AtomGroup = None, positions: np.ndarray[float] = None,
        bonds: np.ndarray[int] = None, dimensions: np.ndarray[float] = None,
        thresholds: np.ndarray[float] = None, masses: np.ndarray[float] = None,
    ) -> np.ndarray[float]:

    r"""
    Locally unwraps the positions of molecules at the edge of the
    simulation box.

    Parameters
    ----------
    group : `MDAnalysis.AtomGroup`, optional
        Atom group. If not provided, the atom positions, bonds, and
        dimensions must be provided in `positions`, `bonds`, and
        `dimensions`, respectively.

    positions : `numpy.ndarray`, optional
        Atom positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    bonds : `numpy.ndarray`, optional
        Pairs of all bonded atom indices with respect to `positions`.

        **Shape**: :math:`(N_\mathrm{bonds},\,2)`.

    dimensions : `numpy.ndarray`, optional
        System dimensions and, optionally, angles.

        **Shape**: :math:`(3,)` or :math:`(6,)`.

        **Reference unit**: :math:`\mathrm{nm}` (lengths) and
        :math:`^\circ` (angles).

    thresholds : `numpy.ndarray`, optional
        Maximum distances in each direction an atom can move before it
        is considered to have crossed a boundary.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    masses : `numpy.ndarray`, optional
        Atom masses. If not specified, all atoms are assumed to have the
        same mass.

        **Shape**: :math:`(N,)`.

        **Reference unit**: :math:`\mathrm{amu}`.

    Returns
    -------
    positions : `numpy.ndarray`
        Unwrapped atom positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.
    """

    if group is not None:
        positions = group.positions
        for fragment in group.fragments:

            # Get locally unwrapped atom positions for fragment
            fragment_positions = make_whole(fragment)

            # Find fragment atoms that are actually in the AtomGroup
            fragment_mask = np.in1d(fragment.ix, group.ix, assume_unique=True)

            # Get corresponding indices for the AtomGroup positions array
            mask = np.in1d(group.ix, fragment.ix[fragment_mask],
                           assume_unique=True)

            # Update the AtomGroup positions
            positions[mask] = fragment_positions[fragment_mask]

        return positions

    elif positions is not None:
        if bonds is None:
            emsg = "Bond information must be specified in 'bonds'."
            raise ValueError(emsg)
        if dimensions is None:
            emsg = "System dimensions must be specified in 'dimensions'."
            raise ValueError(emsg)
        elif len(dimensions) == 3:
            dimensions = np.concatenate((dimensions, (90, 90, 90)))
        if thresholds is None:
            thresholds = dimensions[:3] / 2

        # Get graph of connected atoms
        bond_pairs = {}
        for a, b in bonds:
            bond_pairs.setdefault(a, []).append(b)
            bond_pairs.setdefault(b, []).append(a)

        # Determine indices of connected atoms in each molecule
        molecules = find_connected_nodes(bond_pairs)

        # Specify which atoms have had their positions updated
        done = {m[0] for m in molecules}
        todo = set(range(len(positions))).difference(done)

        # Upwrap atom positions using that of the nearest bonded atom
        # that has already been unwrapped
        while len(todo) > 0:
            for particle_index in todo:
                for bonded_index in bond_pairs[particle_index]:
                    if bonded_index in done:
                        positions[particle_index] = (
                            positions[bonded_index]
                            + minimize_vectors(positions[particle_index]
                                               - positions[bonded_index],
                                               dimensions)
                        )
                        done.add(particle_index)
                        break
            todo -= done

        if masses is None:
            wmsg = ("No masses specified. All atoms are assumed to "
                    "have a mass of 1.")
            warnings.warn(wmsg)
            masses = np.ones(len(positions))
        elif len(masses) == len(molecules):
            masses = np.concatenate(masses)
        elif len(masses) != len(positions):
            emsg = ("The number of masses must be equal to the number "
                    "of atoms or the number of molecules.")
            raise ValueError(emsg)

        # Recenter molecules using their centers of mass
        for molecule_indices in molecules:
            com = center_of_mass(positions=positions[molecule_indices],
                                 masses=masses[molecule_indices])
            positions[molecule_indices] \
                += wrap(com, dimensions[:3], in_place=False) - com

        return positions
    else:
        raise ValueError("Either 'group' or 'positions' must be specified.")

def wrap(
        positions: np.ndarray[float], dimensions: np.ndarray[float], *,
        in_place: bool = True) -> np.ndarray[float]:

    r"""
    Wraps particle positions back into the primary simulation cell.

    Parameters
    ----------
    positions : `numpy.ndarray`
        Particle positions.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    dimensions : `numpy.ndarray`
        System dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    in_place : `bool`, keyword-only, default: :code:`False`
        Determines whether the input array is modified in-place.

    Returns
    -------
    positions : `numpy.ndarray`
        Wrapped particle positions. Only returned if
        :code:`in_place=False`.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.
    """

    wrap_indices = (positions < 0) | (positions > dimensions)
    if in_place:
        positions[wrap_indices] -= (
            np.floor(positions / dimensions) * dimensions
        )[wrap_indices]
    else:
        positions = positions.copy()
        positions[wrap_indices] -= (
            np.floor(positions / dimensions) * dimensions
        )[wrap_indices]
        return positions


"""
Simulation trajectory analysis
==============================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides a variety of classes for analyzing simulation
trajectories.
"""

from . import (base, electrostatics, polymer, profile, structure,
               thermodynamics, transport)

__all__ = ["base", "electrostatics", "polymer", "profile", "structure",
           "thermodynamics", "transport"]


import pathlib
import sys

import MDAnalysis as mda
from MDAnalysis.tests.datafiles import waterDCD, waterPSF
from MDAnalysis.analysis.lineardensity import LinearDensity
import numpy as np

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.analysis import profile # noqa: E402

universe = mda.Universe(waterPSF, waterDCD)

def test_class_density_profile():

    """
    The test cases are adapted from the "Computing mass and charge
    density on each axis" page from the MDAnalysis User Guide
    (https://userguide.mdanalysis.org/stable/examples/analysis/volumetric/linear_density.html).
    """

    density = LinearDensity(universe.atoms, grouping="residues").run()
    density_profile = profile.DensityProfile(
        universe.atoms, "residues", n_bins=200
    ).run()
    parallel_density_profile = profile.DensityProfile(
        universe.atoms, "residues", n_bins=200, parallel=True
    ).run()

    for i, axis in enumerate("xyz"):

        number_density = (
            0.602214076 * getattr(density.results, axis).mass_density
            / universe.residues.masses[0]
        )
        charge_density = (0.602214076
                          * getattr(density.results, axis).charge_density)

        # TEST CASE 1: Number density profiles
        assert(np.allclose(number_density,
                           density_profile.results.number_densities[i]))
        assert(np.allclose(number_density,
                           parallel_density_profile.results.number_densities[i]))

        # TEST CASE 2: Charge density profiles
        assert(np.allclose(charge_density,
                           density_profile.results.charge_densities[i]))
        assert(np.allclose(charge_density,
                           parallel_density_profile.results.charge_densities[i]))


"""
Electrostatics
==============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains classes to quantify electrostatic properties, such
as the instantaneous dipole moment.
"""

from numbers import Real
from typing import Union

import MDAnalysis as mda
from MDAnalysis.lib.mdamath import make_whole
import numpy as np

from .base import DynamicAnalysisBase
from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm.topology import unwrap
from ..algorithm.unit import strip_unit

if FOUND_OPENMM:
    from openmm import unit

def calculate_relative_permittivity(
        M: np.ndarray[float], temperature: float, volume: float, *,
        reduced: bool = False) -> float:

    r"""
    Calculates the relative permittivity (or static dielectric constant)
    :math:`\varepsilon_\mathrm{r}` of a medium using the instantaneous
    dipole moments :math:`\mathbf{M}(t)`.

    The dipole moment fluctuation formula [1]_ relates the relative
    permittivity to the dipole moment via

    .. math::

       \varepsilon_\mathrm{r}=1+\frac{\overline{\langle\mathbf{M}^2\rangle
       -\langle\mathbf{M}\rangle^2}}{3\varepsilon_0 Vk_\mathrm{B}T}

    where the angular brackets :math:`\langle\,\cdot\,\rangle` denote
    the ensemble average, the overline signifies the spatial average,
    :math:`\varepsilon_0` is the vacuum permittivity,
    :math:`k_\mathrm{B}` is the Boltzmann constant, and :math:`T` is
    the system temperature.

    .. note::

       If residues (molecules) in your system have net charges, the
       dipole moments must be made position-independent by subtracting
       the product of the net charge and the center of mass or geometry.

    Parameters
    ----------
    M : array-like
        Instantaneous dipole moments over :math:`N_t` frames.

        **Shape**: :math:`(N_t,\,3)`.

        **Reference unit**: :math:`\mathrm{e\cdotÅ}`.

    temperature : `float`
        System temperature :math:`T`.

        .. note::

           If :code:`reduced=True`, `temperature` should be equal to the
           energy scale. When the Lennard-Jones potential is used, it
           generally means that :math:`T^* = 1`, or `temperature=1`.

        **Reference unit**: :math:`\mathrm{K}`.

    volume : `float`
        System volume :math:`V`.

        **Reference unit**: :math:`\mathrm{Å^3}`.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units.

    Returns
    -------
    dielectric : `float`
        Relative permittivity (or static dielectric constant).

    References
    ----------
    .. [1] Neumann, M. Dipole Moment Fluctuation Formulas in Computer
       Simulations of Polar Systems. *Molecular Physics* **1983**,
       *50* (4), 841–858. https://doi.org/10.1080/00268978300102721.
    """

    if reduced:
        return (1 + 4 * np.pi * (M ** 2 - M.mean(axis=0) ** 2).mean()
                    / (volume.mean() * temperature))
    else:
        M *= ureg.elementary_charge * ureg.angstrom
        temperature *= ureg.kelvin
        volume *= ureg.angstrom ** 3
        return (1 + (M ** 2 - M.mean(axis=0) ** 2).mean()
                / (ureg.vacuum_permittivity * volume.mean()
                   * ureg.boltzmann_constant * temperature)).magnitude

class DipoleMoment(DynamicAnalysisBase):

    r"""
    Serial and parallel implementations to calculate the instantaneous
    dipole moment vectors :math:`\mathbf{M}(t)`.

    For a system with :math:`N` atoms or molecules, the dipole moment is
    given by

    .. math::

       \mathbf{M}=\sum_i^{N}q_i\mathbf{z}_i

    where :math:`q_i` and :math:`\mathbf{z}_i` are the charge and
    position of entity :math:`i`.

    The dipole moment can be used to estimate the relative permittivity
    (or static dielectric constant) via the dipole moment fluctuation
    formula [1]_:

    .. math::

       \varepsilon_\mathrm{r}=1+\frac{\overline{\langle\mathbf{M}^2\rangle
       -\langle\mathbf{M}\rangle^2}}{3\varepsilon_0 Vk_\mathrm{B}T}

    where the angular brackets :math:`\langle\,\cdot\,\rangle` denote
    the ensemble average, the overline signifies the spatial average,
    :math:`\varepsilon_0` is the vacuum permittivity,
    :math:`k_\mathrm{B}` is the Boltzmann constant, and :math:`T` is
    the system temperature.

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Group(s) of atoms for which the dipole moments are calculated.

    charges : array-like, keyword-only, optional
        Charge information for the atoms in the :math:`N_\mathrm{g}`
        groups in `groups`. If not provided, it should be available in
        and will be retrieved from the main
        :class:`MDAnalysis.core.universe.Universe` object.

        **Shape**: :math:`(N_\mathrm{g},)` array of real numbers or
        :math:`(N_i,)` arrays, where :math:`N_i` is the number of atoms
        in group :math:`i`.

        **Reference unit**: :math:`\mathrm{e}`.

    dimensions : array-like, keyword-only, optional
        System dimensions. If the
        :class:`MDAnalysis.core.universe.Universe` object that the
        groups in `groups` belong to does not contain dimensionality
        information, provide it here. Affected by `scales`.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    scales : array-like, keyword-only, optional
        Scaling factors for each system dimension. If an `int` is
        provided, the same value is used for all axes.

        **Shape**: :math:`(3,)`.

    average : `bool`, keyword-only, default: :code:`False`
        Determines whether the dipole moment vectors and volumes are
        time-averaged.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units. Only affects
        :meth:`calculate_relative_permittivity` calls.

    neutralize : `bool`, keyword-only, default: :code:`False`
        Specifies whether the net charge is subtracted at the center of
        mass for molecules with net charges. Must be enabled if your
        system contains molecules with net charges and you want to
        calculate the relative permittivity, but should be disabled if
        you are only interested in the instantaneous dipole moment.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether progress is printed to the console.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for :code:`results.times`, call
        :code:`results.units["results.times"]`.

    results.times : `numpy.ndarray`
        Changes in time :math:`t-t_0`.

        **Shape**: :math:`(N_t,)`.

        **Reference unit**: :math:`\mathrm{ps}`.

    results.dipoles : `numpy.ndarray`
        Instantaneous dipole moment vectors :math:`\mathbf{M}`.

        **Shape**: :math:`(N_t,\,N_\mathrm{g},\,3)`.

        **Reference unit**: :math:`\mathrm{e\cdotÅ}`.

    results.volumes : `numpy.ndarray`
        System volumes :math:`V`.

        **Shape**: :math:`(N_t,)`.

        **Reference unit**: :math:`\mathrm{Å^3}`.

    results.dielectric : `numpy.ndarray`
        Relative permittivity (or static dielectric constant)
        :math:`\varepsilon_\mathrm{r}` in each dimension.

        **Shape**: :math:`(3,)`.

    References
    ----------
    .. [1] Neumann, M. Dipole Moment Fluctuation Formulas in Computer
       Simulations of Polar Systems. *Molecular Physics* **1983**,
       *50* (4), 841–858. https://doi.org/10.1080/00268978300102721.
    """

    def __init__(
            self, groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
            charges: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
            dimensions: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
            scales: Union[float, tuple[float]] = 1, average: bool = False,
            reduced: bool = False, neutralize: bool = False,
            unwrap: bool = False, parallel: bool = False, verbose: bool = True,
            **kwargs) -> None:

        self._groups = [groups] if isinstance(groups, mda.AtomGroup) else groups
        self._n_groups = len(self._groups)
        self.universe = self._groups[0].universe

        super().__init__(self.universe.trajectory, parallel, verbose, **kwargs)

        if dimensions is not None:
            if len(dimensions) != 3:
                raise ValueError("'dimensions' must have length 3.")
            self._dimensions = np.asarray(strip_unit(dimensions, "angstrom")[0])
        elif self.universe.dimensions is not None:
            self._dimensions = self.universe.dimensions[:3].copy()
        else:
            raise ValueError("No system dimensions found or provided.")

        if isinstance(scales, Real) or (len(scales) == 3
                                        and isinstance(scales[0], Real)):
            self._dimensions *= scales
        else:
            emsg = ("The scaling factor(s) must be provided as a "
                    "floating-point number or in an array with shape (3,). ")
            raise ValueError(emsg)

        # Determine the number of particles in each group and their
        # corresponding indices
        self._Ns = np.fromiter((g.n_atoms for g in self._groups), dtype=int,
                               count=self._n_groups)
        self._N = self._Ns.sum()
        self._slices = []
        index = 0
        for N in self._Ns:
            self._slices.append(slice(index, index + N))
            index += N

        if charges is not None:
            charges = list(charges)
            if len(charges) == self._n_groups:
                for i, (g, q) in enumerate(zip(self._groups, charges)):
                    charges[i] = strip_unit(q, "elementary_charge")[0]
                    if isinstance(charges[i], Real):
                        charges[i] *= np.ones(g.n_atoms)
                    elif g.n_atoms != len(charges[i]):
                        emsg = ("The number of charges in "
                                f"'charges[{i}]' is not equal to the "
                                "number of atoms in the corresponding "
                                "group.")
                        raise ValueError(emsg)
                self._charges = charges
            else:
                emsg = ("The number of group charge arrays is not "
                        "equal to the number of groups.")
                raise ValueError(emsg)
        elif hasattr(self.universe.atoms, "charges"):
            self._charges = [g.charges for g in self._groups]
        else:
            raise ValueError("The topology has no charge information.")

        self._all_neutral = np.allclose(
            self.universe.atoms.total_charge(compound="residues"), 0,
            atol=1e-6
        )
        self._all_included = (sum(g.n_atoms for g in self._groups)
                              == self.universe.atoms.n_atoms)

        self._average = average
        self._reduced = reduced
        self._neutralize = neutralize
        self._unwrap = unwrap
        self._verbose = verbose

    def _prepare(self) -> None:

        if self._unwrap:
            self.universe.trajectory[
                self._sliced_trajectory.frames[0]
                if hasattr(self._sliced_trajectory, "frames")
                else (self.start or 0)
            ]

            # Preallocate arrays to store positions and number of
            # boundary crossings
            self._positions_old = np.empty((self._N, 3))
            for g, s in zip(self._groups, self._slices):
                for f in g.fragments:
                    make_whole(f)
                self._positions_old[s] = g.positions
            self._images = np.zeros((self._N, 3), dtype=int)
            self._thresholds = self._dimensions / 2

            # Store unwrapped particle positions in a shared memory array
            # for parallel analysis
            if self._parallel:
                self._positions = np.empty((self.n_frames, self._N, 3))
                for i, _ in enumerate(self.universe.trajectory[
                        list(self._sliced_trajectory.frames)
                        if hasattr(self._sliced_trajectory, "frames")
                        else slice(self.start, self.stop, self.step)
                    ]):
                    for g, s in zip(self._groups, self._slices):
                        self._positions[i, s] = g.positions

                    unwrap(self._positions[i], self._positions_old,
                           self._dimensions, thresholds=self._thresholds,
                           images=self._images)

                del self._positions_old
                del self._images
                del self._thresholds

        # Preallocate arrays to store results and store reference units
        self.results.dipoles = np.zeros((self.n_frames, self._n_groups, 3))
        self.results.volumes = np.empty(self.n_frames)
        self.results.units = {
            "dipoles": ureg.elementary_charge * ureg.angstrom,
            "volumes": ureg.angstrom ** 3
        }
        if not self._average:
            self.results.times = (self.step * self._trajectory.dt
                                  * np.arange(self.n_frames))
            self.results.units["times"] = ureg.picosecond
        if not self._parallel:
            self._positions = np.empty((self._N, 3))

    def _single_frame(self) -> None:

        for i, (g, s, q) in enumerate(zip(self._groups, self._slices,
                                          self._charges)):
            self._positions[s] = g.positions
            self._positions[0, 0] += self._dimensions[0] / 2
            if self._unwrap:
                unwrap(self._positions[s], self._positions_old[s],
                       self._dimensions, thresholds=self._thresholds,
                       images=self._images[s])
            if self._neutralize:
                q -= q * np.concatenate([r.atoms.masses / r.mass
                                         for r in g.residues])
            self.results.dipoles[self._frame_index, i] \
                += q @ self._positions[s]

        self.results.volumes[self._frame_index] \
            = self.universe.trajectory.ts.volume

    def _single_frame_parallel(
            self, frame: int, index: int
        ) -> tuple[int, float, np.ndarray[float]]:

        self._trajectory[frame]
        results = np.empty((self._n_groups, 3))

        for i, (g, s, q) in enumerate(zip(self._groups, self._slices,
                                          self._charges)):
            positions = (self._positions[index, s] if self._unwrap
                         else g.positions)
            if self._neutralize:
                q -= q * np.concatenate([r.atoms.masses / r.mass
                                         for r in g.residues])
            results[i] = q @ positions

        return index, self._trajectory.ts.volume, results

    def _conclude(self) -> None:

        # Consolidate results from parallel analysis
        if self._parallel:
            self._results = sorted(self._results)
            self.results.volumes = np.fromiter((r[1] for r in self._results),
                                               dtype=float, count=self.n_frames)
            self.results.dipoles = np.stack([r[2] for r in self._results])
        else:
            del self._positions
            if self._unwrap:
                del self._positions_old
                del self._images
                del self._thresholds

        # Average results, if requested
        if self._average:
            self.results.dipoles = self.results.dipoles.mean(axis=0)
            self.results.volumes = self.results.volumes.mean()

    def calculate_relative_permittivity(
            self, temperature: Union[float, "unit.Quantity", Q_]) -> None:

        r"""
        Calculates the relative permittivity (or static dielectric
        constant) :math:`\varepsilon_\mathrm{r}` of a medium using the
        instantaneous dipole moments :math:`\mathbf{M}(t)`.

        Parameters
        ----------
        temperature : `float`, `openmm.unit.Quantity`, or `pint.Quantity`
            System temperature :math:`T`.

            .. note::

               If :code:`reduced=True` was set in the
               :class:`DipoleMoment` constructor, `temperature` should
               be equal to the energy scale. When the Lennard-Jones
               potential is used, it generally means that
               :math:`T^* = 1`, or `temperature=1`.

            **Reference unit**: :math:`\mathrm{K}`.
        """

        if self._average:
            emsg = ("Cannot compute relative permittivity using the"
                    "averaged dipole moment.")
            raise RuntimeError(emsg)
        elif not self._all_neutral and not self._neutralize:
            emsg = ("Cannot compute relative permittivity for a "
                    "non-neutral system or a system with ions unless "
                    "the net charge is subtracted at the center of "
                    "mass of each molecule carrying a net charge.")
            raise RuntimeError(emsg)
        elif not self._all_included:
            emsg = ("Cannot compute relative permittivity when not all"
                    "atoms in the system are accounted for in the "
                    "groups.")
            raise RuntimeError(emsg)
        else:
            temperature, unit_ = strip_unit(temperature, "kelvin")
            if self._reduced and not isinstance(unit_, str):
                emsg = ("'temperature' cannot have units when reduced=True.")
                raise ValueError(emsg)

            dipoles = self.results.dipoles
            if self._n_groups > 1:
                dipoles = dipoles.sum(axis=1)
            self.results.dielectric = calculate_relative_permittivity(
                dipoles, temperature, self.results.volumes.mean(),
                reduced=self._reduced
            )


"""
Molecular structure
===================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains algorithms for computing structural information
for a collection of particles.
"""

from typing import Union

import MDAnalysis as mda
import numpy as np

def center_of_mass(
        group: mda.AtomGroup = None, grouping: str = None, *,
        masses: Union[list[float], np.ndarray[float]] = None,
        positions: Union[list[float], np.ndarray[float]] = None,
        images: np.ndarray[int] = None, dimensions: np.ndarray[float] = None,
        n_groups: int = None, raw: bool = False
    ) -> Union[np.ndarray[float],
               tuple[np.ndarray[float], np.ndarray[float], np.ndarray[float]]]:

    r"""
    Computes the centers of mass :math:`\mathbf{R}_\mathrm{com}` for a
    collection of atoms.

    For a group of :math:`N` atoms with masses :math:`m_i` and positions
    :math:`\mathbf{r}_i`, the center of mass is defined as

    .. math::

       \mathbf{R}_\mathrm{com}=\frac{\sum_{i=1}^Nm_i
       \mathbf{r}_i}{\sum_{i=1}^Nm_i}

    .. note::

       This function supports a wide variety of inputs, depending on
       how the atom information is provided and what should be
       calculated.

       When an :class:`MDAnalysis.core.groups.AtomGroup` object is
       provided in `group`, the atom masses and positions are retrieved
       from it and do not need to be provided in `masses` and
       `positions`, respectively. If the :code:`AtomGroup` abides by the
       standard topological heirarchy, you can specify the desired
       grouping in `grouping` and the appropriate centers of mass will
       be calculated. Otherwise, if and only if the :code:`AtomGroup`
       contains equisized or identical groups corresponding to the
       desired grouping (i.e., the :code:`AtomGroup` has atoms that are
       or can be treated as nonbonded entities or topological groups
       with the same number of but not necessarily identical
       constituents), you can provide the total number of groups in
       `n_groups` and the atom masses and positions will be distributed
       accordingly.

       If the trajectory is not unwrapped, the number of periodic
       boundary crossings (and optionally, the system dimensions if they
       are not embedded in the :code:`AtomGroup`) can be provided in
       `images` (and `dimensions`).

       If the :code:`AtomGroup` does not have the correct structural
       information and the residues or segments do not contain the same
       number of atoms, the *unwrapped* atom masses and positions can
       each be provided directly as a :class:`numpy.ndarray`, list, or
       tuple in `masses` and `positions`, respectively. To calculate the
       overall center of mass, the array-like object holding the masses
       should be one-dimensional, while that containing the positions
       should be two-dimensional. To calculate centers of mass for
       multiple groups, the array-like object holding the masses should
       be two-dimensional (indices: group, atom), while that containing
       the positions should be three-dimensional (indices: group, atom,
       axis). When a list or tuple is used, the inner arrays do not have
       to be homogeneously shaped, thus allowing you to calculate the
       centers of mass for residues or segments with different numbers
       of atoms.

       You may also provide only one of the atom masses or positions, in
       which case the missing information will be retrieved from the
       :code:`AtomGroup` provided in `group`. This is generally not
       recommended since the shapes of the provided and retrieved arrays
       may be incompatible.

    Parameters
    ----------
    group : `MDAnalysis.AtomGroup`, optional
        Collection of atoms to compute the centers of mass for. If not
        provided, the atom masses and posititions must be provided
        directly in `masses` and `positions`, respectively.

    grouping : `str`, optional
        Determines which center of mass is calculated if atom masses and
        positions are retrieved from `group`.

        .. container::

           **Valid values**:

           * :code:`None`: Center of mass of all atoms in `group`.
           * :code:`"residues"`: Centers of mass for each residue or
             molecule in `group`.
           * :code:`"segments"`: Centers of mass for each segment or
             chain in `group`.

    masses : array-like, keyword-only, optional
        Atom masses.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,)` for the overall center of mass,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res})` for the residue
             centers of mass, where :math:`N_\mathrm{res}` is
             the number of residues, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg}` for the segment
             centers of mass, where :math:`N_\mathrm{seg}` is
             the number of segments.

           For groups with different numbers of atoms, the array-like
           object should contain inner lists or tuples holding the
           masses of the atoms in each group.

        **Reference unit**: :math:`\mathrm{g/mol}`.

    positions : array-like, keyword-only, optional
        Atom positions.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,\,3)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,\,3)` for the overall center of mass,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res},\,3)` for the
             residue centers of mass, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg},\,3)` for the
             segment centers of mass.

           For groups with different numbers of atoms, the array-like
           object should contain inner lists or tuples holding the
           positions of the atoms in each group.

        **Reference unit**: :math:`\mathrm{Å}`.

    images : array-like, keyword-only, optional
        Image flags for the atoms. Must be provided to get correct
        results if the trajectory is wrapped.

        **Shape**: Same as `positions`.

    dimensions : `numpy.ndarray`, keyword-only, optional
        System dimensions. Must be provided if `images` is provided and
        `group` is not provided or does not contain the system
        dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    n_groups : `int`, keyword-only, optional
        Number of residues or segments. Must be provided if `group` has
        an irregular topological heirarchy or the `masses` and
        `positions` arrays have the general ungrouped shapes.

    raw : `bool`, keyword-only, default: :code:`False`
        Determines whether atom masses and positions are returned if
        they were retrieved from `group`.

    Returns
    -------
    com : `numpy.ndarray`
        Centers of mass.

        .. container::

           **Shape**:

           * :math:`(3,)` for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},\,3)` for
             :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},\,3)` for
             :code:`grouping="segments"`.

    masses : `numpy.ndarray`
        Atom masses. Only returned if `group` was provided and contains
        equisized or identical groups, and :code:`raw=True`.

        .. container::

           **Shape**:

           * :math:`(N,)` for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res})` for
             :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg})` for
             :code:`grouping="segments"`.

        **Reference unit**: :math:`\mathrm{g/mol}`.

    positions : `numpy.ndarray`
        Unwrapped atom positions. Only returned if `group` was
        provided and contains equisized or identical groups, and
        :code:`raw=True`.

        .. container::

           **Shape**:

           * :math:`(N,\,3)` for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res},\,3)` for
             :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg},\,3)` for
             :code:`grouping="segments"`.

        **Reference unit**: :math:`\mathrm{Å}`.
    """

    # Check whether grouping is valid
    if grouping not in {None, "residues", "segments"}:
        emsg = (f"Invalid grouping: '{grouping}'. Valid options are "
                "None, 'residues', and 'segments'.")
        raise ValueError(emsg)

    # Get system dimensions if image flags are provided
    if images is not None:
        if dimensions is None:
            try:
                dimensions = group.dimensions[:3]
            except (NameError, TypeError):
                emsg = ("Image flags were provided, but no system "
                        "dimensions were provided or found in the "
                        "trajectory.")
                raise ValueError(emsg)

    # Get particle masses and positions from the trajectory, if
    # necessary
    missing = (masses is None, positions is None)
    if any(missing):
        if group is None:
            emsg = ("Either a group of atoms or atom positions and "
                    "masses must be provided.")
            raise ValueError(emsg)

        # Check whether the groups have equal numbers of atoms
        if grouping is None:
            same = True
        else:
            groups = getattr(group, grouping)

            # Calculate and return the centers of mass for different
            # groups here if unwrapping and the mass and position arrays
            # are not needed
            if not (same := all(g.atoms.n_atoms == groups[0].atoms.n_atoms
                            for g in groups)) and images is None and not raw:
                return np.array([g.atoms.center_of_mass() for g in groups])

        # Get and unwrap particle positions, if necessary
        if missing[1]:
            positions = group.positions
            if images is not None:
                positions += images * dimensions[:3]

        # Get particle masses and ensure correct dimensionality, if
        # necessary
        if same:
            if missing[0]:
                masses = group.masses
            if grouping is not None or n_groups is not None:
                shape = (n_groups or getattr(group, f"n_{grouping}"), -1, 3)
                masses = masses.reshape(shape[:-1])
                positions = positions.reshape(shape)
        else:
            if missing[0]:
                masses = [g.atoms.masses for g in groups]
            if missing[1]:
                positions = [positions[g.atoms.ix] for g in groups]
    else:

        # Try to convert arrays to NumPy arrays if they are not already
        # to take advantage of vectorized operations later
        try:
            positions = np.asarray(positions)
            masses = np.asarray(masses)
            if images is not None:
                positions += images
        except ValueError:
            pass
        if type(masses) != type(positions):
            emsg = ("The shapes of the arrays containing the masses "
                    "and positions are incompatible.")
            raise ValueError(emsg)
        if images is not None and type(images) != type(positions):
            emsg = ("The shapes of the arrays containing the positions "
                    "and image flags are incompatible.")
            raise ValueError(emsg)

    # Calculate the centers of mass for the specified grouping
    if isinstance(positions, np.ndarray):

        # Reshape the mass and position arrays based on the specified
        # number of groups
        if n_groups is not None:
            masses = masses.reshape((n_groups, -1))
            positions = positions.reshape((n_groups, -1, 3))

        com = (np.einsum("...a,...ad->...d", masses, positions)
               / masses.sum(axis=-1, keepdims=True))
    elif images is None:
        com = np.array([np.dot(m, p) / m.sum()
                        for m, p in zip(masses, positions)])
    else:
        com = np.array([np.dot(m, p + i * dimensions) / m.sum()
                        for m, p, i in zip(masses, positions, images)])

    # Return raw masses and positions, if desired
    if raw and any(missing):
        return com, masses, positions

    return com

def radius_of_gyration(
        group: mda.AtomGroup = None, grouping: str = None, *,
        positions: Union[list[tuple[tuple[float]]], np.ndarray[float]] = None,
        masses: Union[list[tuple[float]], np.ndarray[float]] = None,
        com: np.ndarray[float] = None, images: np.ndarray[int] = None,
        dimensions: np.ndarray[float] = None, n_groups: int = None,
        components: bool = False) -> Union[float, np.ndarray[float]]:

    r"""
    Computes the radii of gyration :math:`R_\mathrm{g}` for a collection
    of atoms.

    For a group of :math:`N` atoms with masses :math:`m_i` and positions
    :math:`\mathbf{r}_i`, the radius of gyration is defined as

    .. math::

       R_\mathrm{g}=\sqrt{
       \frac{\sum_i^Nm_i\|\mathbf{r}_i-\mathbf{R}_\mathrm{com}\|^2}
       {\sum_i^Nm_i}}

    where :math:`\mathbf{R}_\mathrm{com}` is the center of mass.

    Alternatively, the radii of gyration around the coordinate axes can
    be calculated by only summing the radii components orthogonal to
    each axis. For example, the radius of gyration around the
    :math:`x`-axis is

    .. math::

       R_{\mathrm{g},\,x}=\sqrt{\frac{\sum_i^Nm_i
       \left[(\mathbf{r}_{i,\,y}-\mathbf{R}_{\mathrm{com},\,y})^2
       +(\mathbf{r}_{i,\,z}-\mathbf{R}_{\mathrm{com},\,z})^2\right]}
       {\sum_i^Nm_i}}

    .. note::

       This function supports a wide variety of inputs, depending on
       how the atom information is provided and what should be
       calculated.

       When an :class:`MDAnalysis.core.groups.AtomGroup` object is
       provided in `group`, the atom masses and positions are retrieved
       from it and do not need to be provided in `masses` and
       `positions`, respectively. If the :code:`AtomGroup` abides by the
       standard topological heirarchy, you can specify the desired
       grouping in `grouping` and the appropriate centers of mass will
       be calculated. Otherwise, if and only if the :code:`AtomGroup`
       contains equisized or identical groups corresponding to the
       desired grouping (i.e., the :code:`AtomGroup` has atoms that are
       or can be treated as nonbonded entities or topological groups
       with the same number of but not necessarily identical
       constituents), you can provide the total number of groups in
       `n_groups` and the atom masses and positions will be distributed
       accordingly.

       If the trajectory is not unwrapped, the number of periodic
       boundary crossings (and optionally, the system dimensions if they
       are not embedded in the :code:`AtomGroup`) can be provided in
       `images` (and `dimensions`).

       If the :code:`AtomGroup` does not have the correct structural
       information and the residues or segments do not contain the same
       number of atoms, the *unwrapped* atom masses and positions can
       each be provided directly as a :class:`numpy.ndarray`, list, or
       tuple in `masses` and `positions`, respectively. To calculate the
       overall center of mass, the array-like object holding the masses
       should be one-dimensional, while that containing the positions
       should be two-dimensional. To calculate centers of mass for
       multiple groups, the array-like object holding the masses should
       be two-dimensional (indices: group, atom), while that containing
       the positions should be three-dimensional (indices: group, atom,
       axis). When a list or tuple is used, the inner arrays do not have
       to be homogeneously shaped, thus allowing you to calculate the
       centers of mass for residues or segments with different numbers
       of atoms.

       You may also provide only one of the atom masses or positions, in
       which case the missing information will be retrieved from the
       :code:`AtomGroup` provided in `group`. This is generally not
       recommended since the shapes of the provided and retrieved arrays
       may be incompatible.

    Parameters
    ----------
    group : `MDAnalysis.AtomGroup`, optional
        Collection of atoms to compute the centers of mass for. If not
        provided, the atom masses and posititions must be provided
        directly in `masses` and `positions`, respectively.

    grouping : `str`, optional
        Determines which center of mass is calculated if atom masses and
        positions are retrieved from `group`.

        .. container::

           **Valid values**:

           * :code:`None`: Center of mass of all atoms in `group`.
           * :code:`"residues"`: Centers of mass for each residue or
             molecule in `group`.
           * :code:`"segments"`: Centers of mass for each segment or
             chain in `group`.

    masses : array-like, keyword-only, optional
        Atom masses.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,)` for the overall center of mass,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res})` for the residue
             centers of mass, where :math:`N_\mathrm{res}` is
             the number of residues, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg}` for the segment
             centers of mass, where :math:`N_\mathrm{seg}` is
             the number of segments.

           For groups with different numbers of atoms, the array-like
           object should contain inner lists or tuples holding the
           masses of the atoms in each group.

        **Reference unit**: :math:`\mathrm{g/mol}`.

    positions : array-like, keyword-only, optional
        Atom positions.

        .. container::

           **Shape**:

           The general ungrouped shape is :math:`(N,\,3)`.

           For equisized or identical groups, the :class:`numpy.ndarray`
           object should have shape

           * :math:`(N,\,3)` for the overall center of mass,
           * :math:`(N_\mathrm{res},\,N/N_\mathrm{res},\,3)` for the
             residue centers of mass, or
           * :math:`(N_\mathrm{seg},\,N/N_\mathrm{seg},\,3)` for the
             segment centers of mass.

           For groups with different numbers of atoms, the array-like
           object should contain inner lists or tuples holding the
           positions of the atoms in each group.

        **Reference unit**: :math:`\mathrm{Å}`.

    com : `numpy.ndarray`, keyword-only, optional
        Centers of mass.

        .. container::

           **Shape**:

           * :math:`(3,)` for the overall radius of gyration.
           * :math:`(N_\mathrm{res},\,3)` for the residue radii of
             gyration.
           * :math:`(N_\mathrm{seg},\,3)` for the segment radii of
             gyration.

    images : array-like, keyword-only, optional
        Image flags for the atoms. Must be provided to get correct
        results if the trajectory is wrapped.

        **Shape**: Same as `positions`.

    dimensions : `numpy.ndarray`, keyword-only, optional
        System dimensions. Must be provided if `images` is provided and
        `group` is not provided or does not contain the system
        dimensions.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    n_groups : `int`, keyword-only, optional
        Number of residues or segments. Must be provided if `group` has
        an irregular topological heirarchy or the `masses` and
        `positions` arrays have the general ungrouped shapes.

    components : `bool`, keyword-only, default: :code:`False`
        Specifies whether the components of the radii of gyration are
        calculated and returned instead.

    Returns
    -------
    gyradii : `float` or `numpy.ndarray`
        Radii of gyration.

        .. container::

           **Shape**:

           * Scalar for :code:`grouping=None`.
           * :math:`(N_\mathrm{res},)` for :code:`grouping="residues"`.
           * :math:`(N_\mathrm{seg},)` for :code:`grouping="segments"`.

        **Reference unit**: :math:`\mathrm{Å}`.
    """

    # Check whether grouping is valid
    if grouping not in {None, "residues", "segments"}:
        emsg = (f"Invalid grouping: '{grouping}'. Valid options are "
                "None, 'residues', and 'segments'.")
        raise ValueError(emsg)

    # Get particle masses and positions from the trajectory and the
    # center(s) of mass, if necessary
    missing = (masses is None, positions is None, com is None)
    if any(missing[:2]):
        com, masses, positions = center_of_mass(
            group,
            grouping,
            masses=masses,
            positions=positions,
            images=images,
            dimensions=dimensions,
            n_groups=n_groups,
            raw=True
        )
    elif missing[2]:
        com = center_of_mass(masses=masses, positions=positions,
                             n_groups=n_groups)

    if isinstance(positions, np.ndarray):
        if components:
            cpos = (positions
                    - np.expand_dims(com, axis=positions.ndim - 2)) ** 2

            # Compute the radii of gyration in each direction for
            # equisized or identical groups
            if grouping or n_groups:
                return np.sqrt(
                    np.einsum("ga,gad->gd", masses,
                              np.stack((cpos[:, :, (1, 2)].sum(axis=2),
                                        cpos[:, :, (0, 2)].sum(axis=2),
                                        cpos[:, :, (0, 1)].sum(axis=2)),
                                       axis=2))
                    / masses.sum(axis=1, keepdims=True)
                )

            # Compute the radius of gyration in each direction for all
            # atoms
            return np.sqrt(
                np.dot(
                    masses,
                    np.hstack((cpos[:, (1, 2)].sum(axis=1, keepdims=True),
                               cpos[:, (0, 2)].sum(axis=1, keepdims=True),
                               cpos[:, (0, 1)].sum(axis=1, keepdims=True)))
                ) / masses.sum()
            )

        # Compute the overall radii of gyration for equisized or
        # identical groups
        elif grouping or n_groups:
            return np.sqrt(
                np.einsum("ga,gad->gd", masses,
                          (positions - com[:, None]) ** 2).sum(axis=1)
                / masses.sum(axis=1)
            )

        # Compute the overall radius of gyration for all atoms
        return np.sqrt(np.dot(masses, (positions - com) ** 2).sum()
                       / masses.sum())

    # Compute the radii of gyration in each direction for asymmetric
    # groups
    if components:
        gyradii = np.empty(com.shape)
        for i, (m, p, c) in enumerate(zip(masses, positions, com)):
            cpos = (p - c) ** 2
            gyradii[i] = np.array(
                (np.dot(m, cpos[:, (1, 2)].sum(axis=1)),
                 np.dot(m, cpos[:, (0, 2)].sum(axis=1)),
                 np.dot(m, cpos[:, (0, 1)].sum(axis=1)))
            ) / m.sum()
        return np.sqrt(gyradii)

    # Compute the overall radii of gyration for asymmetric groups
    return np.sqrt(
        [np.einsum("a,ad->d", m, (p - c) ** 2).sum() / m.sum()
         for m, p, c in zip(masses, positions, com)]
    )


"""
Polymeric analysis
==================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains classes to determine the structural
and dynamical properties of polymeric systems.
"""

from typing import Union
import warnings

import MDAnalysis as mda
from MDAnalysis.lib.log import ProgressBar
from MDAnalysis.lib.mdamath import make_whole
import numpy as np
from scipy import optimize, special

from .base import DynamicAnalysisBase
from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm import correlation
from ..algorithm.molecule import center_of_mass, radius_of_gyration
from ..algorithm.topology import unwrap, unwrap_edge
from ..algorithm.unit import strip_unit
from ..fit.exponential import stretched_exp

if FOUND_OPENMM:
    from openmm import unit

def correlation_fft(*args, **kwargs):

    """
    Evaluates the autocorrelation function (ACF) or cross-correlation
    function (CCF) of a time series using fast Fourier transforms (FFT).

    .. note::

       This is an alias function. For more information, see
       :func:`mdcraft.algorithm.correlation.correlation_fft`.
    """

    return correlation.correlation_fft(*args, **kwargs)

def correlation_shift(*args, **kwargs) -> np.ndarray[float]:

    """
    Evaluates the autocorrelation function (ACF) or cross-correlation
    function (CCF) of a time series directly by using sliding windows
    along the time axis.

    .. note::

       This is an alias function. For more information, see
       :func:`mdcraft.algorithm.correlation.correlation_shift`.
    """

    return correlation.correlation_shift(*args, **kwargs)

def calculate_relaxation_time(
        time: np.ndarray[float], acf: np.ndarray[float]) -> float:

    r"""
    Calculates the orientational relaxation time :math:`\tau_\mathrm{r}`
    of a polymer using the end-to-end vector autocorrelation function
    (ACF) time series :math:`C_\mathrm{ee}`.

    A stretched exponential function with :math:`\tau` and :math:`\beta`
    as coefficients,

    .. math::

       C_\mathrm{ee}=\exp{\left[-(t/\tau)^\beta\right]}

    is fitted to the ACF time series, and the relaxation time is estimated
    using

    .. math::

       \tau_\mathrm{r}=\int_0^\infty C_\mathrm{ee}\,dt=\tau\Gamma(1/\beta)

    Parameters
    ----------
    time : `numpy.ndarray`
        Changes in time :math:`t-t_0`.

        **Shape**: :math:`(N_t,)`.

        **Reference unit**: :math:`\textrm{ps}`.

    acf : `numpy.ndarray`
        End-to-end vector ACFs for the :math:`N_\textrm{g}` groups over
        :math:`N_\textrm{b}` blocks of :math:`N_t` trajectory frames
        each.

        **Shape**:
        :math:`(N_\textrm{g},\,N_\textrm{b},\,N_t)`.

    Returns
    -------
    relaxation_time : `float`
        Average orientational relaxation time.

        **Reference unit**: :math:`\textrm{ps}`.
    """

    tau_r, beta = optimize.curve_fit(stretched_exp, time / time[1], acf,
                                     bounds=(0, np.inf))[0]
    return tau_r * time[1] * special.gamma(1 + beta ** -1)

class _PolymerAnalysisBase(DynamicAnalysisBase):

    r"""
    An analysis base object for polymer systems.

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Group(s) of polymers to be analyzed. All polymers in each group
        must have the same chain length.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct segment or residue information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (for coarse-grained polymer
             simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic polymer simulations).

    n_chains : `int` or array-like, optional
        Number of chains in each polymer group. Must be provided if the
        trajectory does not adhere to the standard container heirarchy
        (see Notes).

    n_monomers : `int` or array-like, optional
        Number of monomers in each chain in each polymer group. Must be
        provided if the trajectory does not adhere to the standard
        container heirarchy (see Notes).

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.
    """

    def __init__(
            self, groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
            groupings: Union[str, tuple[str]] = "atoms",
            n_chains: Union[int, tuple[int]] = None,
            n_monomers: Union[int, tuple[int]] = None, *, unwrap: bool = False,
            parallel: bool = False, verbose: bool = True, **kwargs) -> None:

        self._groups = [groups] if isinstance(groups, mda.AtomGroup) else groups
        self.universe = self._groups[0].universe

        super().__init__(self.universe.trajectory, parallel, verbose, **kwargs)

        self._dimensions = self.universe.dimensions
        if self._dimensions is not None:
            self._dimensions = self._dimensions[:3].copy()

        self._n_groups = len(self._groups)
        if isinstance(groupings, str):
            if groupings not in (GROUPINGS := {"atoms", "residues"}):
                emsg = (f"Invalid grouping '{groupings}'. Valid values: "
                        f"{', '.join(GROUPINGS)}.")
                raise ValueError(emsg)
            self._groupings = self._n_groups * [groupings]
        else:
            if self._n_groups != len(groupings):
                emsg = ("The number of grouping values is not equal to "
                        "the number of groups.")
                raise ValueError(emsg)
            for g in groupings:
                if g not in (GROUPINGS := {"atoms", "residues"}):
                    emsg = (f"Invalid grouping '{g}'. Valid values: "
                            f"{', '.join(GROUPINGS)}.")
                    raise ValueError(emsg)
            self._groupings = groupings

        if n_chains is None or n_monomers is None:
            self._internal = True
            self._n_chains = np.empty(self._n_groups, dtype=int)
            self._n_monomers = np.empty_like(self._n_chains)
            for i, g in enumerate(self._groups):
                self._n_chains[i] = g.segments.n_segments
                self._n_monomers[i] = g.n_atoms // self._n_chains[i]
        else:
            self._internal = False
            if isinstance(n_chains, (int, np.integer)):
                self._n_chains = n_chains * np.ones(self._n_groups, dtype=int)
            elif self._n_groups == len(n_chains):
                self._n_chains = n_chains
            else:
                emsg = ("The number of polymer counts is not equal to the "
                        "number of groups.")
                raise ValueError(emsg)
            if isinstance(n_monomers, (int, np.integer)):
                self._n_monomers = n_monomers * np.ones(n_monomers, dtype=int)
            elif self._n_groups == len(n_monomers):
                self._n_monomers = n_monomers
            else:
                emsg = ("The number of chain lengths is not equal to the "
                        "number of groups.")
                raise ValueError(emsg)

        self._unwrap = unwrap
        self._verbose = verbose

class Gyradius(_PolymerAnalysisBase):

    r"""
    Serial and parallel implementations to calculate the radius of
    gyration :math:`R_\mathrm{g}` of a polymer.

    The radius of gyration is used to describe the dimensions of a
    polymer chain, and is defined as

    .. math::

        R_\mathrm{g}=\sqrt{
        \frac{\sum_i^N m_i\|\mathbf{r}_i-\mathbf{R}_\mathrm{com}\|^2}
        {\sum_i^N m_i}}

    where :math:`m_i` and :math:`\mathbf{r}_i` are the mass and
    position, respectively, of particle :math:`i`, and
    :math:`\mathbf{R}_\mathrm{com}` is the center of mass.

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or `array_like`
        Group(s) of polymers to be analyzed. All polymers in each group
        must have the same chain length.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct segment or residue information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (for coarse-grained polymer
             simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic polymer simulations).

    n_chains : `int` or array-like, optional
        Number of chains in each polymer group. Must be provided if the
        trajectory does not adhere to the standard container heirarchy
        (see Notes).

    n_monomers : `int` or array-like, optional
        Number of monomers in each chain in each polymer group. Must be
        provided if the trajectory does not adhere to the standard
        container heirarchy (see Notes).

    components : `bool`, keyword-only, default: :code:`False`
        Specifies whether the components of the radii of gyration are
        calculated and returned instead.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for :code:`results.gyradii`, call
        :code:`results.units["results.gyradii"]`.

    results.gyradii : `numpy.ndarray`
        Radii of gyration for the :math:`N_\textrm{g}` groups over
        :math:`N_t` trajectory frames.

        **Shape**: :math:`(N_\textrm{g},\,N_t)`
        (:code:`components=False`) or :math:`(N_\textrm{g},\,N_t,\,3)`
        (:code:`components=True`).

        **Reference unit**: :math:`\textrm{Å}`.
    """

    def __init__(
            self, groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
            groupings: Union[str, tuple[str]] = "atoms",
            n_chains: Union[int, tuple[int]] = None,
            n_monomers: Union[int, tuple[int]] = None, *,
            components: bool = False, unwrap: bool = False,
            parallel: bool = False, verbose: bool = True, **kwargs):

        super().__init__(groups, groupings, n_chains, n_monomers,
                         unwrap=unwrap, parallel=parallel, verbose=verbose,
                         **kwargs)

        # Determine the number of particles in each group and their
        # corresponding indices
        self._Ns = np.fromiter(
            (M * N_p for M, N_p in zip(self._n_chains, self._n_monomers)),
            dtype=int,
            count=self._n_groups
        )
        self._N = self._Ns.sum()
        self._slices = []
        index = 0
        for N in self._Ns:
            self._slices.append(slice(index, index + N))
            index += N

        self._components = components

    def _prepare(self) -> None:

        if self._unwrap:
            self.universe.trajectory[
                self._sliced_trajectory.frames[0]
                if hasattr(self._sliced_trajectory, "frames")
                else (self.start or 0)
            ]

            # Preallocate arrays to store number of boundary crossings for
            # each particle
            self._positions_old = np.empty((self._N, 3))
            for g, gr, s, M, N_p in zip(self._groups, self._groupings,
                                        self._slices, self._n_chains,
                                        self._n_monomers):
                if self._internal and gr == "residues":
                    for f in g.fragments:
                        make_whole(f)
                    self._positions_old[s] = center_of_mass(g, gr)
                else:
                    positions = unwrap_edge(
                        positions=g.positions,
                        bonds=np.array([(i * N_p + j, i * N_p + j + 1)
                                        for i in range(M) for j in range(N_p - 1)]),
                        dimensions=self._dimensions,
                        masses=g.masses
                    )
                    self._positions_old[s] = (
                        positions if gr == "atoms"
                        else center_of_mass(
                            positions=positions.reshape(M, N_p, -1, 3),
                            masses=g.masses.reshape(M, N_p, -1)
                        )
                    )
            self._images = np.zeros((self._N, 3), dtype=int)
            self._thresholds = self._dimensions / 2

            # Store unwrapped particle positions in a shared memory array
            # for parallel analysis
            if self._parallel:
                self._positions = np.empty((self.n_frames, self._N, 3))
                for i, _ in enumerate(self.universe.trajectory[
                        list(self._sliced_trajectory.frames)
                        if hasattr(self._sliced_trajectory, "frames")
                        else slice(self.start, self.stop, self.step)
                    ]):
                    for g, gr, s in zip(self._groups, self._groupings, self._slices):
                        if self._internal and gr == "residues":
                            self._positions[i, s] = center_of_mass(g, gr)
                        else:
                            self._positions[i, s] = (
                                g.positions if gr == "atoms"
                                else center_of_mass(
                                    positions=g.positions.reshape(M, N_p, -1, 3),
                                    masses=g.masses.reshape(M, N_p, -1)
                                )
                            )

                    unwrap(self._positions[i], self._positions_old,
                           self._dimensions, thresholds=self._thresholds,
                           images=self._images)

                # Clean up memory
                del self._positions_old
                del self._images
                del self._thresholds

        if not self._parallel:
            shape = [self._n_groups, self.n_frames]
            if self._components:
                shape.append(3)
            self.results.gyradii = np.empty(shape)
        self.results.units = {"results.gyradii": ureg.angstrom}

    def _single_frame(self) -> None:

        for i, (g, gr, M, N_p, s) in enumerate(
                zip(self._groups, self._groupings, self._n_chains,
                    self._n_monomers, self._slices)
            ):
            if self._internal and gr == "residues":
                positions = center_of_mass(g, gr)
            else:
                positions = (
                    g.positions if gr == "atoms"
                    else center_of_mass(
                        positions=g.positions.reshape(M, N_p, -1, 3),
                        masses=g.masses.reshape(M, N_p, -1)
                    )
                )
            if self._unwrap:
                unwrap(positions, self._positions_old[s], self._dimensions,
                       thresholds=self._thresholds, images=self._images[s])

            self.results.gyradii[i, self._frame_index] \
                = radius_of_gyration(
                    grouping="segments",
                    positions=positions.reshape((M, N_p, 3)),
                    masses=g.masses.reshape((M, N_p)),
                    components=self._components
                ).mean(axis=0)

    def _single_frame_parallel(
            self, frame: int, index: int) -> tuple[int, np.ndarray[float]]:

        self._trajectory[frame]
        shape = [self._n_groups]
        if self._components:
            shape.append(3)
        results = np.empty(shape)

        for i, (g, gr, M, N_p, s) in enumerate(
                zip(self._groups, self._groupings, self._n_chains,
                    self._n_monomers, self._slices)
            ):
            if self._unwrap:
                positions = self._positions[index, s]
            elif self._internal and gr == "residues":
                positions = center_of_mass(g, gr)
            else:
                positions = (
                    g.positions if gr == "atoms"
                    else center_of_mass(
                        positions=g.positions.reshape(M, N_p, -1, 3),
                        masses=g.masses.reshape(M, N_p, -1)
                    )
                )

            results[i] = radius_of_gyration(
                grouping="segments",
                positions=positions.reshape((M, N_p, 3)),
                masses=g.masses.reshape((M, N_p)),
                components=self._components
            ).mean(axis=0)

        return index, results

    def _conclude(self) -> None:

        # Consolidate parallel results
        if self._parallel:
            self.results.gyradii = np.stack(
                [r[1] for r in sorted(self._results)], axis=1
            )

class EndToEndVector(_PolymerAnalysisBase):

    r"""
    A serial implementation to calculate the end-to-end vector
    autocorrelation function (ACF) :math:`C_\mathrm{ee}(t)` and
    estimate the orientational relaxation time :math:`\tau_\mathrm{r}`
    of a polymer.

    The end-to-end vector ACF is defined as

    .. math::

       C_\mathrm{ee}(t)=\frac{\langle\mathbf{R}_\mathrm{ee}(t)
       \cdot\mathbf{R}_\mathrm{ee}(0)\rangle}
       {\langle\mathbf{R}_\mathrm{ee}^2\rangle}

    where :math:`\mathbf{R}_\mathrm{ee}=\mathbf{r}_N-\mathbf{r}_1`
    is the end-to-end vector.

    The orientational relaxation time can then be estimated by fitting
    a stretched exponential function

    .. math::

       C_\mathrm{ee}=\exp{\left[-(t/\tau)^\beta\right]}

    to the end-to-end vector ACF and evaluating

    .. math::

       \tau_\mathrm{r}=\int_0^\infty C_\mathrm{ee}\,dt
       =\tau\Gamma(\frac{1}{\beta}+1)

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or `array_like`
        Group(s) of polymers to be analyzed. All polymers in each group
        must have the same chain length.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct segment or residue information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (for coarse-grained polymer
             simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic polymer simulations).

    n_chains : `int` or array-like, optional
        Number of chains in each polymer group. Must be provided if the
        trajectory does not adhere to the standard container heirarchy
        (see Notes).

    n_monomers : `int` or array-like, optional
        Number of monomers in each chain in each polymer group. Must be
        provided if the trajectory does not adhere to the standard
        container heirarchy (see Notes).

    n_blocks : `int`, keyword-only, default: :code:`1`
        Number of blocks to split the trajectory into.

    dt : `float` or `openmm.unit.Quantity`, keyword-only, optional
        Time between frames :math:`\Delta t`. While this is normally
        determined from the trajectory, the trajectory may not have the
        correct information if the data is in reduced units. For
        example, if your reduced timestep is :math:`0.01` and you output
        trajectory data every :math:`10,000` timesteps, then
        :math:`\Delta t=100`.

        **Reference unit**: :math:`\mathrm{ps}`.

    fft : `bool`, keyword-only, default: :code:`True`
        Determines whether fast Fourier transforms (FFT) are used to
        evaluate the ACFs.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for :code:`results.times`, call
        :code:`results.units["results.times"]`.

    results.times : `numpy.ndarray`
        Changes in time :math:`t-t_0`.

        **Shape**: :math:`(N_t,)`.

        **Reference unit**: :math:`\textrm{ps}`.

    results.acf : `numpy.ndarray`
        End-to-end vector ACFs for the :math:`N_\textrm{g}` groups over
        :math:`N_\textrm{b}` blocks with :math:`N_t` trajectory frames
        each.

        **Shape**:
        :math:`(N_\textrm{g},\,N_\textrm{b},\,N_t)`.

    results.relaxation_times : `numpy.ndarray`
        Average orientational relaxation times for the
        :math:`N_\textrm{g}` groups over :math:`N_t` trajectory frames
        split into :math:`N_\textrm{b}` blocks.

    Notes
    -----
    In a standard trajectory file, segments (or chains) contain
    residues, and residues contain atoms. This heirarchy must be adhered
    to for this analysis module to function correctly. If your
    trajectory file does not contain the correct segment or residue
    information, provide the number of chains and chain lengths in
    `n_chains` and `n_monomers`, respectively.
    """

    def __init__(
            self, groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
            groupings: Union[str, tuple[str]] = "atoms",
            n_chains: Union[int, tuple[int]] = None,
            n_monomers: Union[int, tuple[int]] = None, *, n_blocks: int = 1,
            dt: Union[float, "unit.Quantity", Q_] = None, fft: bool = True,
            unwrap: bool = False, verbose: bool = True, **kwargs) -> None:

        # Disable parallel support for this analysis class
        if "parallel" in kwargs:
            del kwargs["parallel"]

        super().__init__(groups, groupings, n_chains, n_monomers,
                         unwrap=unwrap, verbose=verbose, **kwargs)

        self._N_chains = self._n_chains.sum()
        self._slices = []
        index = 0
        for N in self._n_chains:
            self._slices.append(slice(index, index + N))
            index += N

        self._n_blocks = n_blocks
        self._dt = strip_unit(dt or self._trajectory.dt, "picosecond")[0]
        self._fft = fft

    def _prepare(self) -> None:

        # Determine number of frames used when the trajectory is split
        # into blocks
        self._n_frames_block = self.n_frames // self._n_blocks
        self._n_frames = self._n_blocks * self._n_frames_block
        extra_frames = self.n_frames - self._n_frames
        if extra_frames > 0:
            wmsg = (f"The trajectory is not divisible into {self._n_blocks:,} "
                    f"blocks, so the last {extra_frames:,} frame(s) will be "
                    "discarded. To maximize performance, set appropriate "
                    "starting and ending frames in run() so that the number "
                    "of frames to be analyzed is divisible by the number of "
                    "blocks.")
            warnings.warn(wmsg)

        self._e2e = np.empty((self.n_frames, self._N_chains, 3))
        if self._unwrap:
            self.universe.trajectory[
                self._sliced_trajectory.frames[0]
                if hasattr(self._sliced_trajectory, "frames")
                else (self.start or 0)
            ]

            # Preallocate arrays to store number of boundary crossings for
            # the first and last monomer in each chain
            self._positions_end_old = np.empty((self._N_chains, 2, 3))
            for g, gr, s, M, N_p in zip(self._groups, self._groupings,
                                        self._slices, self._n_chains,
                                        self._n_monomers):
                if self._internal and gr == "residues":
                    for f in g.fragments:
                        make_whole(f)
                    self._positions_end_old[s] = np.stack(
                        [center_of_mass(s.residues[[0, -1]].atoms, "residues")
                         for s in g.segments]
                    )
                else:
                    positions = unwrap_edge(
                        positions=g.positions,
                        bonds=np.array([(i * N_p + j, i * N_p + j + 1)
                                        for i in range(M) for j in range(N_p - 1)]),
                        dimensions=self._dimensions,
                        masses=g.masses
                    ).reshape(M, N_p, -1, 3)[:, (0, -1)]
                    self._positions_end_old[s] = (
                        positions[:, :, 0] if gr == "atoms"
                        else center_of_mass(
                            positions=positions,
                            masses=g.masses.reshape(M, N_p, -1)[:, (0, -1)]
                        )
                    )
            self._images = np.zeros((self._N_chains, 2, 3), dtype=int)
            self._thresholds = self._dimensions / 2

        self.results.times = self.step * self._dt * np.arange(self._n_frames //
                                                              self._n_blocks)
        self.results.acf = np.empty(
            (self._n_groups, self._n_blocks, self._n_frames_block)
        )
        self.results.units = {"results.times": ureg.picosecond}

    def _single_frame(self) -> None:

        for g, gr, s, M, N_p in zip(self._groups, self._groupings,
                                    self._slices, self._n_chains,
                                    self._n_monomers):
            if self._internal and gr == "residues":
                positions_end = np.stack(
                    center_of_mass(s.residues[[0, -1]].atoms, "residues")
                    for s in g.segments
                )
            else:
                positions_end = g.positions.reshape(M, N_p, -1, 3)[:, (0, -1)]
                positions_end = (
                    positions_end[:, :, 0] if gr == "atoms"
                    else center_of_mass(
                        positions=positions_end,
                        masses=g.masses.reshape(M, N_p, -1)[:, (0, -1)]
                    )
                )
            if self._unwrap:
                unwrap(positions_end, self._positions_end_old[s],
                       self._dimensions, thresholds=self._thresholds,
                       images=self._images[s])
            self._e2e[self._frame_index, s] \
                = np.diff(positions_end, axis=1)[:, 0]

    def _conclude(self) -> None:

        # Clean up memory
        if self._unwrap:
            del self._positions_end_old
            del self._images
            del self._thresholds

        _acf = correlation_fft if self._fft else correlation_shift
        for i, (s, M) in ProgressBar(enumerate(zip(self._slices,
                                                   self._n_chains))):
            self.results.acf[i] = _acf(
                (self._e2e[:, s]
                 / np.linalg.norm(self._e2e[:, s], axis=-1, keepdims=True))
                .reshape(self._n_blocks, -1, M, 3),
                average=True, vector=True
            )

    def calculate_relaxation_time(self) -> None:

        """
        Calculates the orientational relaxation time.
        """

        if not hasattr(self.results, "acf"):
            emsg = ("Call EndToEndVector.run() before "
                    "EndToEndVector.calculate_relaxation_time().")
            raise RuntimeError(emsg)

        self.results.relaxation_times = np.empty((self._n_groups,
                                                  self._n_blocks))
        self.results.units["results.relaxation_times"] = ureg.picosecond

        for i, g in enumerate(self.results.acf):
            for j, acf in enumerate(g):
                valid = np.where(acf >= 0)[0]
                self.results.relaxation_times[i, j] = calculate_relaxation_time(
                    self.results.times[valid], acf[valid]
                )

class SingleChainStructureFactor(DynamicAnalysisBase):

    r"""
    Serial and parallel implementations to calculate the single-chain
    structure factor :math:`S_\mathrm{sc}(q)` of a homopolymer.

    It is defined as

    .. math::

       S_{\mathrm{sc}}(\mathbf{q})=\frac{1}{MN_\mathrm{p}}
       \sum_{m=1}^M\sum_{i,j=1}^{N_\mathrm{p}}\left\langle
       \exp{[i\mathbf{q}\cdot(\mathbf{r}_i-\mathbf{r}_j)]}\right\rangle

    where :math:`M` is the number of chains, :math:`N_\mathrm{p}` is the
    chain length, :math:`\mathbf{q}` is the scattering wavevector, and
    :math:`\mathbf{r}_i` is the position of the :math:`i`-th monomer.

    .. container::

       The single-chain structure factor reveals information about the
       characteristic length scales of the polymer:

       * In the Guinier regime (:math:`qR_g\ll1`),
         :math:`S_{\mathrm{sc}}(q)^{-1}\approx N_\mathrm{p}(1-(qR_g)^2/3)`
         can be used to determine the radius of gyration :math:`R_g`.
       * In the Porod regime (:math:`qR_g\gg1`),
         :math:`S_{\mathrm{sc}}(q)=1` since the only contribution is the
         self-scattering of the monomers.
       * In the intermediate regime, the slope :math:`s` of the log-log
         plot of :math:`S_{\mathrm{sc}}(q)` is related to the scaling
         exponent :math:`\nu` via :math:`\nu=-1/s`.

    Parameters
    ----------
    group : `MDAnalysis.AtomGroup`
        Group of polymers to be analyzed. All polymers in the group
        must have the same chain length.

    grouping : `str`, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct segment or residue information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (for coarse-grained polymer
             simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic polymer simulations).

    n_points : `int`, default: :code:`32`
        Number of points to sample the wavevector space.

    n_chains : `int`, optional
        Number of chains in `group`. Must be provided if the topology
        does not contain segment information.

    n_monomers : `int`, optional
        Number of monomers per chain. Must be provided if the topology
        does not contain segment information.

    dimensions : `numpy.ndarray` or `openmm.unit.Quantity`, optional
        System dimensions. If the
        :class:`MDAnalysis.core.universe.Universe` object that `group`
        belongs to does not contain dimensionality information, provide
        it here.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\textrm{Å}`.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for :code:`results.wavenumbers`, call
        :code:`results.units["results.wavenumbers"]`.

    results.wavenumbers : `numpy.ndarray`
        Unique wavenumbers.

        **Shape**: :math:`(N_q,)`.

        **Reference unit**: :math:`\textrm{Å}^{-1}`.

    results.scsf : `numpy.ndarray`
        Single-chain structure factors for the unique wavenumbers.

        **Shape**: :math:`(N_q,)`.

    Notes
    -----
    In a standard trajectory file, segments (or chains) contain
    residues, and residues contain atoms. This heirarchy must be adhered
    to for this analysis module to function correctly. If your
    trajectory file does not contain the correct segment or residue
    information, provide the number of chains and chain lengths in
    `n_chains` and `n_monomers`, respectively.
    """

    def __init__(
            self, group: mda.AtomGroup, grouping: str = "atoms",
            n_points: int = 32, *, n_chains: int = None, n_monomers: int = None,
            dimensions: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
            unwrap: bool = False, parallel: bool = False, verbose: bool = True,
            **kwargs) -> None:

        self._group = group
        self.universe = group.universe

        super().__init__(self.universe.trajectory, parallel, verbose, **kwargs)

        if dimensions is not None:
            if len(dimensions) != 3:
                raise ValueError("'dimensions' must have length 3.")
            self._dimensions = np.asarray(strip_unit(dimensions, "angstrom")[0])
        elif self.universe.dimensions is not None:
            self._dimensions = self.universe.dimensions[:3].copy()
        else:
            raise ValueError("No system dimensions found or provided.")

        if grouping not in (groupings := {"atoms", "residues"}):
            emsg = (f"Invalid grouping '{grouping}'. Valid values: "
                    f"{', '.join(groupings)}.")
            raise ValueError(emsg)
        self._grouping = grouping

        if n_chains is None or n_monomers is None:
            self._internal = True
            self._n_chains = self._group.segments.n_segments
            self._n_monomers = self._group.n_atoms // self._n_chains
        else:
            self._internal = False
            if isinstance(n_chains, (int, np.integer)):
                self._n_chains = n_chains
            else:
                emsg = ("The number of chains must be specified when "
                        "the universe does not contain segment "
                        "information.")
                raise ValueError(emsg)
            if isinstance(n_monomers, (int, np.integer)):
                self._n_monomers = n_monomers
            else:
                emsg = ("The number of monomers per chain must be "
                        "specified when the universe does not contain "
                        "segment information.")
                raise ValueError(emsg)

        self._n_points = n_points
        self._wavevectors = np.stack(
            np.meshgrid(
                *[2 * np.pi * np.arange(self._n_points) / L
                 for L in self._dimensions]
            ), -1
        ).reshape(-1, 3)
        self._wavenumbers = np.linalg.norm(self._wavevectors, axis=1)

        self._unwrap = unwrap
        self._verbose = verbose

    def _prepare(self) -> None:

        # Unwrap particle positions, if necessary
        if self._unwrap:
            self.universe.trajectory[
                self._sliced_trajectory.frames[0]
                if hasattr(self._sliced_trajectory, "frames")
                else (self.start or 0)
            ]

            if self._internal and self._grouping == "residues":
                self._positions_old = center_of_mass(self._group,
                                                     self._grouping)
            else:
                self._positions_old = (
                    self._group.positions if self._grouping == "atoms"
                    else center_of_mass(
                        positions=self._group.positions.reshape(
                            self._n_chains, self._n_monomers, -1, 3
                        ),
                        masses=self._group.masses.reshape(
                            self._n_chains, self._n_monomers, -1
                        )
                    )
                )
            self._images = np.zeros(self._positions_old.shape, dtype=int)
            self._thresholds = self._dimensions / 2

        # Determine the unique wavenumbers
        self.results.wavenumbers = np.unique(self._wavenumbers.round(11))
        self.results.units = {"results.wavenumbers": ureg.angstrom ** -1}

        # Store unwrapped atom positions in a shared memory array for
        # parallel analysis
        if self._parallel:
            self._positions = np.empty(
                (self.n_frames, self._n_chains * self._n_monomers, 3)
            )

            # Store particle positions in a shared memory array
            for i, _ in enumerate(self.universe.trajectory[
                    list(self._sliced_trajectory.frames)
                    if hasattr(self._sliced_trajectory, "frames")
                    else slice(self.start, self.stop, self.step)
                ]):

                # Store atom or center-of-mass positions in the current frame
                if self._internal and self._grouping == "residues":
                    self._positions[i] = center_of_mass(self._group,
                                                        self._grouping)
                else:
                    self._positions[i] = (
                        self._group.positions if self._grouping == "atoms"
                        else center_of_mass(
                            positions=self._group.positions.reshape(
                                self._n_chains, self._n_monomers, -1, 3
                            ),
                            masses=self._group.masses.reshape(
                                self._n_chains, self._n_monomers, -1
                            )
                        )
                    )

                # Unwrap particle positions if necessary
                if self._unwrap:
                    unwrap(self._positions[i], self._positions_old,
                           self._dimensions, thresholds=self._thresholds,
                           images=self._images)

            # Clean up memory
            del self._positions_old
            del self._thresholds
            del self._images

        # Preallocate arrays to store results
        else:
            self.results.scsf = np.zeros(len(self._wavevectors))

    def _single_frame(self) -> None:

        # Get atom or center-of-mass positions in the current frame
        if self._internal and self._grouping == "residues":
            positions = center_of_mass(self._group, self._grouping)
        else:
            positions = (
                self._group.positions if self._grouping == "atoms"
                else center_of_mass(
                    positions=self._group.positions.reshape(
                        self._n_chains, self._n_monomers, -1, 3
                    ),
                    masses=self._group.masses.reshape(
                        self._n_chains, self._n_monomers, -1
                    )
                )
            )

        # Unwrap particle positions if necessary
        if self._unwrap:
            unwrap(positions, self._positions_old, self._dimensions,
                   thresholds=self._thresholds, images=self._images)

        # Calculate single-chain structure factor contributions
        for chain in positions.reshape((self._n_chains, self._n_monomers, 3)):
            arg = np.einsum("ij,kj->ki", self._wavevectors, chain)
            self.results.scsf += (np.sin(arg).sum(axis=0) ** 2
                                  + np.cos(arg).sum(axis=0) ** 2)

    def _single_frame_parallel(
            self, frame: int, index: int) -> np.ndarray[float]:

        # Compute the single-chain structure factor by squaring the
        # cosine and sine terms and adding them together
        scsf = np.zeros(len(self._wavevectors), dtype=float)
        for chain in self._positions[index].reshape((self._n_chains, -1, 3)):
            arg = np.einsum("ij,kj->ki", self._wavevectors, chain)
            scsf += np.sin(arg).sum(axis=0) ** 2 + np.cos(arg).sum(axis=0) ** 2
        return scsf

    def _conclude(self) -> None:

        # Tally single-chain structure factor for each wavevector over
        # all frames and normalize by the number of particles and timesteps
        if self._parallel:
            self.results.scsf = np.vstack(self._results).sum(axis=0)

        # Normalize the single-chain structure factor by
        # dividing by the total number of monomers and timesteps
        self.results.scsf /= self._n_chains * self._n_monomers * self.n_frames

        # Flatten the array for each combination by combining
        # values sharing the same wavevector magnitude
        self.results.scsf = np.fromiter(
            (self.results.scsf[np.isclose(q, self._wavenumbers)].mean()
             for q in self.results.wavenumbers),
            dtype=float,
            count=len(self.results.wavenumbers)
        )


"""
Linear profiles
===============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains classes to quantify properties along axes, such as
density profiles.
"""

import logging
from numbers import Real
from typing import Union
import warnings

import MDAnalysis as mda
import numpy as np
from scipy import integrate, sparse

from .base import DynamicAnalysisBase
from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm.molecule import center_of_mass
from ..algorithm.topology import unwrap, wrap
from ..algorithm.unit import strip_unit

if FOUND_OPENMM:
    from openmm import unit

def calculate_potential_profile(
        bins: np.ndarray[float], charge_density: np.ndarray[float],
        L: float, dielectric: float = 1, *, sigma_q: float = None,
        dV: float = None, threshold: float = 1e-5, V0: float = 0,
        method: str = "integral", pbc: bool = False, reduced: bool = False
    ) -> np.ndarray[float]:

    r"""
    Calculates the potential profile :math:`\Psi(z)` using the charge
    density profile by numerically solving Poisson's equation for
    electrostatics.

    Poisson's equation is given by

    .. math::

       \varepsilon_0\varepsilon_\mathrm{r}\nabla^2\Psi(z)=-\rho_q(z)

    where :math:`\varepsilon_0` is the vacuum permittivity,
    :math:`\varepsilon_\mathrm{r}` is the relative permittivity,
    :math:`\rho_q` is the charge density, and :math:`\Psi` is the
    potential.

    The boundary conditions (BCs) are

    .. math::

       \left.\frac{\partial\Psi}{\partial z}\right\rvert_{z=0}&
       =-\frac{\sigma_q}{\varepsilon_0\varepsilon_\mathrm{r}}\\
       \left.\Psi\right\rvert_{z=0}&=\Psi_0

    The first BC is used to ensure that the electric field in the bulk
    of the solution is zero, while the second BC is used to set the
    potential on the left electrode.

    Poisson's equation can be evaluated by using the trapezoidal rule
    to numerically integrate the charge density profile twice:

    1. Integrate the charge density profile.
    2. Apply the first BC by adding :math:`\sigma_q` to all points.
    3. Integrate the profile from Step 2.
    4. Divide by :math:`-\varepsilon_0\varepsilon_\mathrm{r}`.
    5. Apply the second BC by adding :math:`\Psi_0` to all points.

    This method is fast but requires many histogram bins to accurately
    resolve the potential profile.

    Alternatively, Poisson's equation can be written as a system of
    linear equations

    .. math::

       A\mathbf{\Psi}=\mathbf{b}

    using second-order finite differences. :math:`A` is a tridiagonal
    matrix and :math:`b` is a vector containing the charge density
    profile, with boundary conditions applied in the first and last
    rows.

    The inner equations are given by

    .. math::

       \Psi_i^{''}=\frac{\Psi_{i-1}-2\Psi_i+\Psi_{i+1}}{h^2}
       =-\frac{1}{\varepsilon_0\varepsilon_\mathrm{r}}\rho_{q,\,i}

    where :math:`i` is the bin index and :math:`h` is the bin width.

    In the case of periodic boundary conditions, the first and last
    equations are given by

    .. math::

       \Psi_0^{''}&=\frac{\Psi_{N-1}-2\Psi_0+\Psi_1}{h^2}
       =-\frac{1}{\varepsilon_0\varepsilon_\mathrm{r}}\rho_{q,\,0}\\
       \Psi_{N-1}^{''}&=\frac{\Psi_{N-2}-2\Psi_{N-1}+\Psi_0}{h^2}
       =-\frac{1}{\varepsilon_0\varepsilon_\mathrm{r}}\rho_{q,\,N-1}

    When the system has slab geometry, the boundary conditions are
    implemented via

    .. math::

       \Psi_0&=\Psi_0\\
       \Psi_0^\prime&=\frac{-3\Psi_0+4\Psi_1-\Psi_2}{2h}
       =-\frac{\sigma_q}{\varepsilon_0\varepsilon_\mathrm{r}}

    This method is slower but can be more accurate even with fewer
    histogram bins for bulk systems with periodic boundary conditions.

    Parameters
    ----------
    bins : array-like
        Histogram bin centers corresponding to the charge density
        profile in `charge_density`.

        **Shape**: :math:`(N_\mathrm{bins},)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    charge_density : array-like
        Array containing the charge density profile.

        **Shape**: :math:`(N_\mathrm{bins},)`.

        **Reference unit**: :math:`\mathrm{e/Å}^{-3}`.

    L : `float`
        System size in the dimension that `bins` and `charge_density`
        were calculated in.

        **Reference unit**: :math:`\mathrm{Å}`.

    dielectric : `float`, default: :code:`1`
        Relative permittivity or static dielectric constant
        :math:`\varepsilon_\mathrm{r}`.

    sigma_q : `float`, keyword-only, optional
        Total surface charge density :math:`\sigma_q`. Used to
        ensure that the electric field in the bulk of the solution
        is zero. If not provided, it is determined using `dV` and
        the charge density profile, or the average value in the
        center of the integrated charge density profile if
        :code:`method="integral"`.

        .. note::

           This value should be negative if the potential difference
           is positive and vice versa.

        **Reference unit**: :math:`\mathrm{e/Å^2}`.

    dV : `float`, keyword-only, optional
        Potential difference :math:`\Delta\Psi` across the system
        dimension specified in `axis`. Has no effect if `sigma_q` is
        provided since this value is used solely to calculate
        `sigma_q`.

        **Reference unit**: :math:`\mathrm{V}`.

    threshold : `float`, keyword-only, default: :code:`1e-5`
        Threshold for determining the plateau region of the first
        integral of the charge density profile to calculate
        `sigma_q`. Has no effect if `sigma_q` is provided or if
        `sigma_q` can be calculated using `dV` and `dielectric`.

    V0 : `float`, keyword-only, default: :code:`0`
        Potential :math:`\Psi_0` at the left boundary.

        **Reference unit**: :math:`\mathrm{V}`.

    method : `str`, keyword-only, default: :code:`"integral"`
        Method to use to calculate the potential profile.

        **Valid values**: :code:`"integral"`, :code:`"matrix"`.

    pbc : `bool`, keyword-only, default: :code:`False`
        Specifies whether the axis has periodic boundary conditions.
        Only used when :code:`method="matrix"`.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units.

    Returns
    -------
    potential : `numpy.ndarray`
        Potential profile :math:`\Psi(z)`.

        **Shape**: :math:`(N_\mathrm{bins},)`.

        **Reference unit**: :math:`\mathrm{V}`.
    """

    if len(bins) != len(charge_density):
        emsg = ("'bins' and 'charge_density' arrays must have the same "
                "length.")
        raise ValueError(emsg)

    CONVERSION_FACTOR = 4 * np.pi if reduced else (
        1 * ureg.elementary_charge / (ureg.vacuum_permittivity * ureg.angstrom)
    ).m_as(ureg.volt)

    # Calculate surface charge density for system with perfectly
    # conducting boundaries
    if sigma_q is None and dV is not None:
        sigma_q = (integrate.trapezoid(bins * charge_density, bins)
                   - dielectric * dV / CONVERSION_FACTOR) / L

    if method == "integral":
        potential = integrate.cumulative_trapezoid(charge_density, bins,
                                                   initial=0)

        if sigma_q is None:
            wmsg = ("No surface charge density information. The value "
                    "will be extracted from the integrated charge "
                    "density profile, which may be inaccurate due to "
                    "numerical errors.")
            warnings.warn(wmsg)

            # Get surface charge density from the integrated charge
            # density profile
            cut_indices = np.where(
                np.diff(np.abs(np.gradient(potential)) < threshold)
            )[0] + 1
            if len(cut_indices) == 0:
                logging.warning(
                    "No bulk plateau region found in the charge "
                    "density profile. The average value over the "
                    "entire profile will be used."
                )
                sigma_q = potential.mean()
            else:
                target_index = len(potential) // 2
                sigma_q = potential[
                    cut_indices[cut_indices <= target_index][-1]:
                    cut_indices[cut_indices >= target_index][0]
                ].mean()

        return (
            -CONVERSION_FACTOR
            * integrate.cumulative_trapezoid(potential + sigma_q, bins,
                                             initial=V0) / dielectric
        )

    elif method == "matrix":
        if sigma_q is None:
            emsg = ("No surface charge density information. Either "
                    "'sigma_q' or 'dV' must be provided when "
                    "method='matrix'.")
            raise ValueError(emsg)

        h = bins[1] - bins[0]
        if not np.allclose(np.diff(bins), h):
            raise ValueError("'bins' must be uniformly spaced.")

        # Set up matrix and load vector for second-order finite
        # difference method
        N = len(bins)
        A = sparse.diags((1, -2, 1), (-1, 0, 1), shape=(N, N), format="csc")
        b = charge_density.copy()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore",
                                  category=sparse.SparseEfficiencyWarning)
            if pbc:
                A[0, -1] = A[-1, 0] = 1
                b *= -CONVERSION_FACTOR * h ** 2 / dielectric
                psi = np.empty_like(b)
                psi[1:] = sparse.linalg.spsolve(A[1:, 1:], b[1:])
                psi[0] = psi[-1]
                return psi
            else:
                A[0, :3] = -1.5, 2, -0.5
                A[-1, 0] = 1
                A[-1, -2:] = 0
                b[0] = -CONVERSION_FACTOR * h * sigma_q / dielectric
                b[1:-1] *= -CONVERSION_FACTOR * h ** 2 / dielectric
                b[-1] = 0
                return sparse.linalg.spsolve(A, b)

class DensityProfile(DynamicAnalysisBase):

    """
    Serial and parallel implementations to calculate the number and
    charge density profiles :math:`\\rho_i(z)` and :math:`\\rho_q(z)` of
    a system along the specified axes.

    The microscopic number density profile of species :math:`i` in a
    constant-volume system is calculated by binning particle positions
    along an axis :math:`z` using

    .. math::

       \\rho_i(z)=\\frac{V}{N_\\mathrm{bin}}\\left\\langle
       \\sum_\\alpha\\delta(z-z_\\alpha)\\right\\rangle

    where :math:`V` is the system volume and :math:`N_\\mathrm{bins}` is
    the number of bins. The angular brackets denote an ensemble average.

    If the species carry charges, the charge density profile can be
    obtained using

    .. math::

       \\rho_q(z)=\\sum_i z_ie\\rho_i(z)

    where :math:`z_i` is the charge number of species :math:`i` and
    :math:`e` is the elementary charge.

    With the charge density profile, the potential profile can be
    computed by numerically solving Poisson's equation for
    electrostatics:

    .. math::

       \\varepsilon_0\\varepsilon_\\mathrm{r}\\nabla^2\\Psi(z)=-\\rho_q(z)

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Groups of atoms for which density profiles are calculated.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           If the desired grouping is not :code:`"atoms"`,

           * the trajectory file should have segments (or chains)
             containing residues (or molecules) and residues containing
             atoms, and

           * residues and segments should be locally unwrapped at the
             simulation box edges, if not already, using
             :class:`MDAnalysis.transformations.wrap.unwrap`,
             :meth:`MDAnalysis.core.groups.AtomGroup.unwrap`, or
             :func:`MDAnalysis.lib.mdamath.make_whole`.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (generally or for
             coarse-grained simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic simulations).
           * :code:`"segments"`: Segments' centers of mass (for
             atomistic polymer simulations).

    axes : `int`, `str`, or array-like, default: :code:`"xyz"`
        Axes along which to compute the density profiles.

        .. container::

           **Examples**:

           * :code:`2` for the :math:`z`-direction.
           * :code:`"xy"` for the :math:`x`- and :math:`y`-directions.
           * :code:`(0, 1)` for the :math:`x`- and :math:`y`-directions.

    n_bins : `int` or array-like
        Number of bins for each axis. If an `int` is provided, the same
        value is used for all axes. If :code:`parallel=True`, the number
        of bins in all axes must be the same.

    charges : array-like, keyword-only, optional
        Charge numbers :math:`z_i` for the specified `groupings` in the
        :math:`N_\\mathrm{g}` `groups`. If not provided, it will be
        retrieved from the main
        :class:`MDAnalysis.core.universe.Universe` object if available.

        .. note::

           Depending on the grouping for a specific group, all atoms,
           residues, or segments should have the same charge since the
           charge density profile for the group would not make sense
           otherwise. If this condition does not hold, change how the
           particles are grouped in `grouping` such that all entities
           share the same charge.

        **Shape**: :math:`(N_\\mathrm{g})`.

        **Reference unit**: :math:`\\mathrm{e}`.

    dimensions : array-like, keyword-only, optional
        System dimensions. If the
        :class:`MDAnalysis.core.universe.Universe` object that the
        groups in `groups` belong to does not contain dimensionality
        information, provide it here. Affected by `scales`.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\\mathrm{Å}`.

    dt : `float`, `openmm.unit.Quantity`, or `pint.Quantity`, \
    keyword-only, optional
        Time between frames :math:`\\Delta t`. While this is normally
        determined from the trajectory, the trajectory may not have the
        correct information if the data is in reduced units. For
        example, if your reduced timestep is :math:`0.01` and you output
        trajectory data every :math:`10,000` timesteps, then
        :math:`\\Delta t = 100`.

        **Reference unit**: :math:`\\mathrm{ps}`.

    scales : array-like, keyword-only, optional
        Scaling factors for each system dimension. If an `int` is
        provided, the same value is used for all axes.

        **Shape**: :math:`(3,)`.

    average : `bool`, keyword-only, default: :code:`True`
        Determines whether the density profiles are averaged over the
        specified frames.

    recenter : `int`, `list`, `MDAnalysis.AtomGroup`, or `tuple`, \
    keyword-only, optional
        Constrains the center of mass of an atom group by adjusting the
        particle coordinates every analysis frame. Either specify an
        :class:`MDAnalysis.core.groups.AtomGroup`, its index within
        `groups`, or a tuple containing the aforementioned information
        and the fixed center of mass coordinates, in that order. To
        avoid recentering in a specific dimension, set the coordinate to
        :code:`numpy.nan`. If the center of mass is not specified, the
        center of the simulation box is used.

        **Shape**: :math:`(3,)` for the fixed center of mass.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units. Affects
        `results.number_densities`, `results.charge_densities`, etc.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the simulation system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for :code:`results.bins`, call
        :code:`results.units["results.bins"]`.

    results.times : `numpy.ndarray`
        Times at which the density profiles are calculated.

        **Shape**: :math:`(N_\\mathrm{frames},)`.

        **Reference unit**: :math:`\\mathrm{ps}`.

    results.bins : `list`
        Bin centers corresponding to the density profiles in each
        dimension.

        **Shape**: :math:`(N_\\mathrm{axes},)` list of
        :math:`(N_\\mathrm{bins},)` arrays.

        **Reference unit**: :math:`\\mathrm{Å}`.

    results.number_densities : `list`
        Number density profiles.

        **Shape**: :math:`(N_\\mathrm{axes},)` list of
        :math:`(N_\\mathrm{bins},)` arrays.

        **Reference unit**: :math:`\\mathrm{Å}^{-3}`.

    results.charge_densities : `list`
        Charge density profiles, if charge information is available.

        **Shape**: :math:`(N_\\mathrm{axes},)` list of
        :math:`(N_\\mathrm{bins},)` arrays.

        **Reference unit**: :math:`\\mathrm{e/Å}^{-3}`.

    results.potentials : `dict`
        Potential profiles, if charge information is available, with
        the key being the axis index (e.g., :code:`1` for the :math:`z`-
        direction if :code:`axes="yz"`). Only available after running
        :meth:`calculate_potential_profile`.

        **Shape**: :math:`(N_\\mathrm{bins},)` for the potential
        profiles.

        **Reference unit**: :math:`\\mathrm{V}`.
    """

    def __init__(
            self, groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
            groupings: Union[str, tuple[str]] = "atoms",
            axes: Union[int, str, tuple[Union[int, str]]] = "xyz",
            n_bins: Union[int, tuple[int]] = 201, *,
            charges: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
            dimensions: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
            dt: Union[float, "unit.Quantity", Q_] = None,
            scales: Union[float, tuple[float]] = 1, average: bool = True,
            recenter: Union[mda.AtomGroup, int,
                            tuple[Union[mda.AtomGroup, int],
                                  np.ndarray[float]]] = None,
            reduced: bool = False, parallel: bool = False,
            verbose: bool = True, **kwargs) -> None:

        self._groups = [groups] if isinstance(groups, mda.AtomGroup) else groups
        self.universe = self._groups[0].universe

        super().__init__(self.universe.trajectory, parallel, verbose, **kwargs)

        self._n_groups = len(self._groups)
        if isinstance(groupings, str):
            if groupings not in (GROUPINGS := {"atoms", "residues",
                                               "segments"}):
                emsg = (f"Invalid grouping '{groupings}'. Valid values: "
                        f"{', '.join(GROUPINGS)}.")
                raise ValueError(emsg)
            self._groupings = self._n_groups * [groupings]
        else:
            if self._n_groups != len(groupings):
                emsg = ("The number of grouping values is not equal to "
                        "the number of groups.")
                raise ValueError(emsg)
            for g in groupings:
                if g not in (GROUPINGS := {"atoms", "residues", "segments"}):
                    emsg = (f"Invalid grouping '{g}'. Valid values: "
                            f"{', '.join(GROUPINGS)}.")
                    raise ValueError(emsg)
            self._groupings = groupings

        if isinstance(axes, int):
            self._axes = np.array((axes,), dtype=int)
        else:
            self._axes = np.fromiter(
                (ord(a.lower()) - 120 if isinstance(a, str) else a
                 for a in axes),
                count=len(axes),
                dtype=int
            )

        if isinstance(n_bins, int):
            self._n_bins = n_bins * np.ones(self._axes.shape, dtype=int)
        elif not isinstance(n_bins, str):
            if len(n_bins) == len(self._axes):
                if parallel and np.any(n_bins != n_bins[0]):
                    emsg = ("All axes must use the same number of bins "
                            "when parallel=True.")
                    raise ValueError(emsg)
                self._n_bins = n_bins
            else:
                emsg = ("The dimension of the array of bin counts is "
                        "incompatible with the number of axes to "
                        "calculate density profiles along.")
                raise ValueError(emsg)
        else:
            emsg = ("The specified bin counts must be an integer or an "
                    "iterable object.")
            raise ValueError(emsg)

        if charges is not None:
            if len(charges) != self._n_groups:
                emsg = ("The number of group charges is not equal to "
                        "the number of groups.")
                raise ValueError(emsg)
            charges, unit_ = strip_unit(charges, "elementary_charge")
            if reduced and not isinstance(unit_, str):
                emsg = "'charges' cannot have units when reduced=True."
                raise TypeError(emsg)
            self._charges = np.asarray(charges)
        elif hasattr(self.universe.atoms, "charges"):
            self._charges = np.empty(self._n_groups)
            for i, (g, gr) in enumerate(zip(self._groups, self._groupings)):
                qs = getattr(g, gr).charges
                if not np.allclose((q := qs[0]), qs):
                    self._charges = None
                    wmsg = (f"Not all {gr} in group {i} share the same "
                            "charge. The charge density profile will "
                            "be calculated.")
                    warnings.warn(wmsg)
                    break
                self._charges[i] = q
        else:
            self._charges = None

        if dimensions is not None:
            if len(dimensions) != 3:
                raise ValueError("'dimensions' must have length 3.")
            self._dimensions = np.asarray(
                strip_unit(dimensions, "angstrom")[0]
            )
        elif self.universe.dimensions is not None:
            self._dimensions = self.universe.dimensions[:3].copy()
        else:
            raise ValueError("No system dimensions found or provided.")

        if isinstance(scales, Real) or (len(scales) == 3
                                        and isinstance(scales[0], Real)):
            self._dimensions *= scales
        else:
            emsg = ("The scaling factor(s) must be provided as a "
                    "floating-point number or in an array with shape "
                    "(3,).")
            raise ValueError(emsg)

        self._dt, unit_ = strip_unit(dt or self._trajectory.dt, "picosecond")
        if reduced and not isinstance(unit_,  str):
            raise TypeError("'dt' cannot have units when reduced=True.")

        if recenter is None:
            self._recenter = recenter
        else:
            if isinstance(recenter, (int, mda.AtomGroup)):
                recenter_group = recenter
                recenter_position = self._dimensions / 2
            elif isinstance(recenter, tuple) and len(recenter) == 2:
                recenter_group, recenter_position = recenter
                recenter_position = np.asarray(recenter_position)
            else:
                emsg = ("Invalid value passed to 'recenter'. The argument "
                        "must either be a MDAnalysis.AtomGroup, its index "
                        "in 'groups', multiple groups/indices, or a tuple "
                        "containing the aforementioned information and a "
                        "specified center of mass, in that order.")
                raise ValueError(emsg)
            if isinstance(recenter_group, int) \
                    and not 0 <= recenter_group < self._n_groups:
                raise ValueError("Invalid group index passed to 'recenter'.")
            elif isinstance(recenter_group, mda.AtomGroup):
                try:
                    recenter_group = self._groups.index(recenter_group)
                except ValueError:
                    emsg = ("The specified AtomGroup in 'recenter' is not "
                            "in 'groups'.")
                    raise ValueError(emsg)
            self._recenter = (recenter_group, recenter_position)

        # Determine the number of particles in each group and their
        # corresponding indices
        self._Ns = np.fromiter(
            (getattr(a, f"n_{g}")
             for (a, g) in zip(self._groups, self._groupings)),
            dtype=int,
            count=self._n_groups
        )
        self._N = self._Ns.sum()
        self._slices = []
        index = 0
        for N in self._Ns:
            self._slices.append(slice(index, index + N))
            index += N

        self._average = average
        self._reduced = reduced
        self._verbose = verbose

    def _prepare(self) -> None:

        self.results.bins = [
            np.linspace(
                self._dimensions[a] / (2 * self._n_bins[i]),
                self._dimensions[a]
                - self._dimensions[a] / (2 * self._n_bins[i]),
                self._n_bins[i]
            ) for i, a in enumerate(self._axes)
        ]

        if self._recenter is not None:

            # Navigate to first frame in analysis
            self.universe.trajectory[
                self._sliced_trajectory.frames[0]
                if hasattr(self._sliced_trajectory, "frames")
                else (self.start or 0)
            ]

            # Preallocate arrays to store number of boundary crossings
            # for each particle
            self._positions_old = np.empty((self._N, 3))
            for g, gr, s in zip(self._groups, self._groupings, self._slices):
                self._positions_old[s] = (g.positions if gr == "atoms"
                                          else center_of_mass(g, gr))
            self._images = np.zeros((self._N, 3), dtype=int)
            self._thresholds = self._dimensions / 2

            # [Parallel] Store unwrapped particle positions in a shared
            # memory array for parallel analysis
            if self._parallel:
                self._positions = np.empty((self.n_frames, self._N, 3))

                # Store atom or center-of-mass positions in the current
                # frame
                for i, _ in enumerate(self._sliced_trajectory):
                    for g, gr, s in zip(self._groups, self._groupings,
                                        self._slices):
                        self._positions[i, s] = (g.positions if gr == "atoms"
                                                 else center_of_mass(g, gr))

                    # Globally unwrap positions for correct center of
                    # mass calculation
                    unwrap(self._positions[i], self._positions_old,
                           self._dimensions, thresholds=self._thresholds,
                           images=self._images)

                    # Calculate difference in center of mass and
                    # shift particle positions by the difference
                    scom = center_of_mass(
                        positions=self._positions
                                  [i, self._slices[self._recenter[0]]],
                        masses=getattr(
                            self._groups[self._recenter[0]],
                            self._groupings[self._recenter[0]]
                        ).masses
                    )
                    self._positions[i] -= np.fromiter(
                        (0 if np.isnan(cx) else sx - cx
                         for sx, cx in zip(scom, self._recenter[1])),
                        dtype=float,
                        count=3
                    )

                # Wrap positions back into the simulation box so that they
                # belong to a histogram bin
                wrap(self._positions, self._dimensions)

        # [Serial] Preallocate arrays to hold atom positions for a
        # given frame so that it doesn't have to be recreated each
        # frame, and number density profiles
        if not self._parallel:
            self._positions = np.empty((self._N, 3))
            shape = [self._n_groups]
            if not self._average:
                shape.append(self.n_frames)
            self.results.number_densities = [np.zeros((*shape, n))
                                             for n in self._n_bins]

        # Store time information
        if not self._average:
            if hasattr(self._sliced_trajectory, "frames"):
                self.results.times \
                    = np.asarray(self._sliced_trajectory.frames) * self._dt
            else:
                self.results.times = self._dt * np.arange(
                    self._sliced_trajectory.start,
                    self._sliced_trajectory.stop,
                    self._sliced_trajectory.step
                )

        # Store reference units
        self.results.units = {
            "results.bins": ureg.angstrom,
            "results.number_densities": ureg.angstrom ** -3
        }
        if self._charges is not None:
            self.results.charge_densities = [None for _ in self._axes]
            self.results.units["results.charge_densities"] = (
                ureg.elementary_charge / ureg.angstrom ** 3
            )

    def _single_frame(self):

        # Store atom or center-of-mass positions in the current frame
        for g, gr, s in zip(self._groups, self._groupings, self._slices):
            self._positions[s] = (g.positions if gr == "atoms"
                                  else center_of_mass(g, gr))

        if self._recenter is not None:

            # Globally unwrap positions for correct center of mass
            # calculation
            unwrap(self._positions, self._positions_old, self._dimensions,
                   thresholds=self._thresholds, images=self._images)

            # Calculate difference in centers of mass and shift particle
            # positions by the difference
            scom = center_of_mass(
                positions=self._positions[self._slices[self._recenter[0]]],
                masses=getattr(self._groups[self._recenter[0]],
                               self._groupings[self._recenter[0]]).masses
            )
            self._positions -= np.fromiter(
                (0 if np.isnan(cx) else sx - cx
                 for sx, cx in zip(scom, self._recenter[1])),
                dtype=float,
                count=3
            )

        # Wrap positions back into the simulation box so that they
        # belong to a histogram bin
        wrap(self._positions, self._dimensions)

        # Compute and tally the bin counts for the current positions
        for i, (gr, s) in enumerate(zip(self._groupings, self._slices)):
            for a, (axis, n_bins) in enumerate(zip(self._axes, self._n_bins)):
                if self._average:
                    self.results.number_densities[a][i] += np.histogram(
                        self._positions[s, axis], n_bins,
                        (0, self._dimensions[axis])
                    )[0]
                else:
                    self.results.number_densities[a][i, self._frame_index] \
                        = np.histogram(self._positions[s, axis], n_bins,
                                       (0, self._dimensions[axis]))[0]

    def _single_frame_parallel(
            self, frame: int, index: int) -> tuple[int, np.ndarray[float]]:

        # Set current trajectory frame
        self._trajectory[frame]

        # Preallocate array to hold bin counts for the current frame
        results = np.empty((len(self._axes), self._n_groups, self._n_bins[0]))

        if self._recenter is None:
            positions = np.empty((self._N, 3))
            for g, gr, s in zip(self._groups, self._groupings, self._slices):
                positions[s] = (g.positions if gr == "atoms"
                                else center_of_mass(g, gr))
            wrap(positions, self._dimensions)
        else:
            positions = self._positions[index]

        # Compute and tally the bin counts for the current positions
        for i, (gr, s) in enumerate(zip(self._groupings, self._slices)):
            for a, axis in enumerate(self._axes):
                results[a, i] = np.histogram(
                    positions[s, axis], self._n_bins[0], (0, self._dimensions[axis])
                )[0]

        return index, results

    def _conclude(self):

        # Consolidate parallel results and clean up memory by deleting
        # arrays that will not be reused
        if self._parallel:
            if self._recenter is not None:
                del self._positions
            self.results.number_densities = np.stack(
                [r[1] for r in sorted(self._results)], axis=1
            )
            if self._average:
                self.results.number_densities \
                    = self.results.number_densities.sum(axis=1)
        else:
            del self._positions
        if self._recenter is not None:
            del self._positions_old, self._images, self._thresholds

        V = np.prod(self._dimensions)
        for a in range(len(self._axes)):
            denom = self._n_bins[a] / V
            if self._average:
                denom /= self.n_frames
            self.results.number_densities[a] *= denom

            if self._charges is not None:
                self.results.charge_densities[a] = np.einsum(
                    "g,...gb->...b",
                    self._charges,
                    self.results.number_densities[a]
                )

    def calculate_potential_profile(
            self, dielectric: float, axis: Union[int, str], *,
            sigma_q: Union[float, "unit.Quantity", Q_] = None,
            dV: Union[float, "unit.Quantity", Q_] = None,
            threshold: float = 1e-5, V0: Union[float, "unit.Quantity", Q_] = 0,
            method: str = "integral", pbc: bool = False) -> None:

        """
        Calculates the average potential profile in the given dimension
        using the charge density profile by numerically solving Poisson's
        equation for electrostatics.

        Parameters
        ----------
        dielectric : `float`
            Relative permittivity or dielectric constant
            :math:`\\varepsilon_\\mathrm{r}`.

        axis : `int` or `str`
            Axis along which to compute the potential profiles.

            .. container::

               **Examples**:

               * :code:`2` for the :math:`z`-direction.
               * :code:`"x"` for the :math:`x`-direction.

        sigma_q : `float`, `openmm.unit.Quantity`, or `pint.Quantity`, \
        keyword-only, optional
            Total surface charge density :math:`\\sigma_q`. Used to
            ensure that the electric field in the bulk of the solution
            is zero. If not provided, it is determined using `dV` and
            the charge density profile, or the average value in the
            center of the integrated charge density profile.

            **Reference unit**: :math:`\\mathrm{e/Å^2}`.

        dV : `float`, `openmm.unit.Quantity`, or `pint.Quantity`, \
        keyword-only, optional
            Potential difference :math:`\\Delta\\Psi` across the system
            dimension specified in `axis`. Has no effect if `sigma_q` is
            provided since this value is used solely to calculate
            `sigma_q`.

            **Reference unit**: :math:`\\mathrm{V}`.

        threshold : `float`, keyword-only, default: :code:`1e-5`
            Threshold for determining the plateau region of the first
            integral of the charge density profile to calculate
            `sigma_q`. Has no effect if `sigma_q` is provided, or if
            `sigma_q` can be calculated using `dV` and `dielectric`.

        V0 : `float`, `openmm.unit.Quantity`, or `pint.Quantity`, \
        keyword-only, default: :code:`0`
            Potential :math:`\\Psi_0` at the left boundary.

            **Reference unit**: :math:`\\mathrm{V}`.

        method : `str`, keyword-only, default: :code:`"integral"`
            Method to use to calculate the potential profile.

            **Valid values**: :code:`"integral"`, :code:`"matrix"`.

        pbc : `bool`, keyword-only, default: :code:`False`
            Specifies whether the system has periodic boundary conditions.
            Only used when :code:`method="matrix"`.
        """

        if not hasattr(self.results, "charge_densities"):
            emsg = ("Either call run() before calculate_potential_profile() "
                    "or provide charge information when initializing "
                    "the DensityProfile object.")
            raise RuntimeError(emsg)

        if not hasattr(self.results, "potentials"):
            self.results.potentials = {}
            self.results.units["results.potentials"] = ureg.volt

        if isinstance(axis, str):
            axis = ord(axis.lower()) - 120
        index = np.where(self._axes == axis)[0][0]

        if sigma_q is not None:
            sigma_q, unit_ = strip_unit(sigma_q,
                                        "elementary_charge/angstrom**2")
            if self._reduced and not isinstance(unit_, str):
                emsg = "'sigma_q' cannot have units when reduced=True."
                raise ValueError(emsg)

        if dV is not None:
            dV, unit_ = strip_unit(dV, "volt")
            if self._reduced and not isinstance(unit_, str):
                emsg = "'dV' cannot have units when reduced=True."
                raise ValueError(emsg)

        if V0 is not None:
            V0, unit_ = strip_unit(V0, "volt")
            if self._reduced and not isinstance(unit_, str):
                emsg = "'V0' cannot have units when reduced=True."
                raise ValueError(emsg)

        charge_density = self.results.charge_densities[index]
        if charge_density.ndim == 3:
            charge_density = charge_density.mean(axis=1)
        self.results.potentials[index] = calculate_potential_profile(
            self.results.bins[index],
            charge_density,
            self._dimensions[axis],
            dielectric,
            sigma_q=sigma_q,
            dV=dV,
            threshold=threshold,
            V0=V0,
            method=method,
            pbc=pbc,
            reduced=self._reduced
        )


"""
Unit manipulation
=================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains helper functions for unit manipulation and reduction.
"""

from numbers import Number
from typing import Union

import numpy as np

from .. import FOUND_OPENMM, Q_, ureg

if FOUND_OPENMM:
    from openmm import unit
    from ..openmm.unit import VACUUM_PERMITTIVITY

def get_scaling_factors(
        bases: dict[str, Union["unit.Quantity", Q_]],
        other: dict[str, list] = {}) -> dict[str, Union["unit.Quantity", Q_]]:

    r"""
    Evaluates scaling factors for reduced units.

    Parameters
    ----------
    bases : `dict`
        Fundamental quantities: molar mass (:math:`m`), length
        (:math:`\sigma`), and energy (:math:`\epsilon`).

        .. container::

           **Format**:

           .. code::

              {
                "mass": <openmm.unit.Quantity> | <pint.Quantity>,
                "length": <openmm.unit.Quantity> | <pint.Quantity>,
                "energy": <openmm.unit.Quantity> | <pint.Quantity>
              }

        **Reference units**: :math:`\mathrm{g/mol}`, :math:`\mathrm{nm}`,
        and :math:`\mathrm{kJ}`.

    other : `dict`, optional
        Other scaling factors to compute. The key should be the name of
        the scaling factor, and the value should contain `tuple`
        objects with the names of bases or default scaling factors and
        their powers.

        **Example**:
        :code:`{"diffusivity": (("length", 2), ("time", -1))}`.

    Returns
    -------
    scales : `dict`
        Scaling factors.
    """

    for name, params in other.items():
        factor = 1
        for base, power in params:
            factor *= bases[base] ** power
        bases[name] = factor

    return bases

def get_lj_scaling_factors(
        bases: dict[str, Union["unit.Quantity", Q_]],
        other: dict[str, list] = {}) -> dict[str, Union["unit.Quantity", Q_]]:

    r"""
    Evaluates scaling factors for Lennard-Jones reduced units.

    By default, the following scaling factors are calculated:

    ===========================  ===================================================
    Quantity (`dict` key)        Expression
    ===========================  ===================================================
    :code:`"molar_energy"`       :math:`N_\mathrm{A}\epsilon`
    :code:`"time"`               :math:`\sqrt{m\sigma^2/\epsilon}`
    :code:`"velocity"`           :math:`\sigma/\tau`
    :code:`"force"`              :math:`\epsilon/\sigma`
    :code:`"temperature"`        :math:`\epsilon/k_\mathrm{B}T`
    :code:`"pressure"`           :math:`\epsilon/\sigma^3`
    :code:`"dynamic_viscosity"`  :math:`\epsilon\tau/\sigma^3`
    :code:`"charge"`             :math:`\sqrt{4\pi\varepsilon_0\sigma\epsilon}`
    :code:`"dipole"`             :math:`\sqrt{4\pi\varepsilon_0\sigma^3\epsilon}`
    :code:`"electric_field"`     :math:`\sqrt{\epsilon/(4\pi\varepsilon_0\sigma^3)}`
    :code:`"mass_density"`       :math:`m/\sigma^3`
    ===========================  ===================================================

    Parameters
    ----------
    bases : `dict`
        Fundamental quantities: molar mass (:math:`m`), length
        (:math:`\sigma`), and energy (:math:`\epsilon`).

        .. container::

           **Format**:

           .. code::

              {
                "mass": <openmm.unit.Quantity> | <pint.Quantity>,
                "length": <openmm.unit.Quantity> | <pint.Quantity>,
                "energy": <openmm.unit.Quantity> | <pint.Quantity>
              }

        **Reference units**: :math:`\mathrm{g/mol}`, :math:`\mathrm{nm}`,
        and :math:`\mathrm{kJ}`.

    other : `dict`, optional
        Other scaling factors to compute. The key should be the name of
        the scaling factor, and the value should contain `tuple`
        objects with the names of bases or default scaling factors and
        their powers.

        **Example**:
        :code:`{"diffusivity": (("length", 2), ("time", -1))}`.

    Returns
    -------
    scales : `dict`
        Scaling factors.
    """

    if bases["mass"].__module__ == "pint":
        avogadro_constant = ureg.avogadro_constant
        boltzmann_constant = ureg.boltzmann_constant
        bases["molar_energy"] = bases["energy"] * avogadro_constant
        bases["time"] = np.sqrt(
            bases["mass"] * bases["length"] ** 2 / bases["molar_energy"]
        ).to(ureg.picosecond)
        bases["charge"] = np.sqrt(
            4 * np.pi * ureg.vacuum_permittivity * bases["length"] * bases["energy"]
        ).to(ureg.elementary_charge)
    else:
        avogadro_constant = unit.AVOGADRO_CONSTANT_NA
        boltzmann_constant = unit.BOLTZMANN_CONSTANT_kB
        bases["molar_energy"] = bases["energy"] * avogadro_constant
        bases["time"] = (
            bases["mass"] * bases["length"] ** 2 / bases["molar_energy"]
        ).sqrt().in_units_of(unit.picosecond)
        bases["charge"] = (
            4 * np.pi * VACUUM_PERMITTIVITY * bases["length"] * bases["energy"]
        ).sqrt().in_units_of(unit.elementary_charge)

    # Define the default scaling factors
    bases["velocity"] = bases["length"] / bases["time"]
    bases["force"] = bases["molar_energy"] / bases["length"]
    bases["temperature"] = bases["energy"] / boltzmann_constant
    bases["pressure"] = bases["energy"] / bases["length"] ** 3
    bases["dynamic_viscosity"] = bases["pressure"] * bases["time"]
    bases["dipole"] = bases["length"] * bases["charge"]
    bases["electric_field"] = bases["force"] / bases["charge"]
    bases["mass_density"] = bases["mass"] / (bases["length"] ** 3
                                             * avogadro_constant)

    return get_scaling_factors(bases, other)

def strip_unit(
        value: Union[Number, "unit.Quantity", Q_],
        unit_: Union[str, "unit.Unit", ureg.Unit] = None
    ) -> tuple[Number, Union[None, "unit.Unit", ureg.Unit]]:

    """
    Strips the unit from an :obj:`openmm.unit.quantity.Quantity` or
    :obj:`pint.Quantity` object.

    Parameters
    ----------
    value : `numbers.Number`, `openmm.unit.Quantity`, or `pint.Quantity`
        Physical quantity for which to get the magnitude of in the
        unit specified in `unit_`.

    unit_ : `str`, `openmm.unit.Unit`, or `pint.Unit`, optional
        Unit to convert to. If not specified, the original unit is used.

    Returns
    -------
    value : `numbers.Number`
        Magnitude of the physical quantity in the specified unit.

    unit_ : `openmm.unit.Unit` or `pint.Unit`
        Unit of the physical quantity.

    Examples
    --------
    For any quantity other than a :obj:`openmm.unit.quantity.Quantity`
    or :obj:`pint.Quantity` object, the raw quantity and user-specified
    unit are returned.

    >>> strip_unit(90.0, "deg")
    (90.0, 'deg')
    >>> strip_unit(90.0, ureg.degree)
    (90.0, <Unit('degree')>)

    If no target unit is specified, the magnitude and original unit of
    the quantity are returned.

    >>> strip_unit(1.380649e-23 * ureg.joule * ureg.kelvin ** -1)
    (1.380649e-23, <Unit('joule / kelvin')>)

    If a target unit using the same module as the quantity is specified,
    the quantity is first converted to the target unit, if necessary,
    before its magnitude and unit are returned.

    >>> g = 9.80665 * ureg.meter / ureg.second ** 2
    >>> strip_unit(g, "meter/second**2")
    (9.80665, <Unit('meter / second ** 2')>)
    >>> strip_unit(g, ureg.foot / ureg.second ** 2)
    (32.17404855643044, <Unit('foot / second ** 2')>)

    If a target unit using a different module than the quantity is
    specified, the quantity is converted to the specified target unit in
    the new module, if necessary, before its magnitude and unit are
    returned.

    >>> strip_unit(8.205736608095969e-05 * unit.meter ** 3 * unit.atmosphere
    ...            / (unit.kelvin * unit.mole),
    ...            ureg.joule / (ureg.kelvin * ureg.mole))
    (8.31446261815324, <Unit('joule / kelvin / mole')>)
    >>> strip_unit(8.205736608095969e-05 * ureg.meter ** 3 * ureg.atmosphere
    ...            / (ureg.kelvin * ureg.mole),
    ...            unit.joule / (unit.kelvin * unit.mole))
    (8.31446261815324, Unit({BaseUnit(..., name="kelvin", ...): -1.0, ...}))
    """

    if isinstance(value, Q_):

        # No target unit (unit_) specified; return Pint magnitude and
        # unit (unit__)
        if unit_ is None:
            unit__ = value.units
            value = value.magnitude
        else:

            # Convert OpenMM target unit (unit__ = unit_) to Pint unit
            # (unit_) and return unit__
            if getattr(unit_, "__module__", None) == "openmm.unit.unit":
                unit_, unit__ = ureg.Unit(""), unit_
                for u, p in unit__.iter_base_or_scaled_units():
                    unit_ *= ureg.Unit(u.name.replace(" ", "_")) ** p

            # Convert str or Pint target unit (unit_) to Pint unit
            # (unit__)
            else:
                unit__ = ureg.Unit(unit_)

            # Get magnitude of Pint quantity in Pint unit (unit_)
            value = value.m_as(unit_)

    elif getattr(value, "__module__", None) == "openmm.unit.quantity":
        swap = False

        # No target unit (unit_) specified; return OpenMM magnitude and
        # unit (unit__)
        if unit_ is None:
            unit_ = unit__ = value.unit

        # Store OpenMM target unit (unit_) in return value (unit__)
        elif getattr(unit_, "__module__", None) == "openmm.unit.unit":
            unit__ = unit_
        else:

            # Determine whether unit_ and unit__ need to be swapped at
            # the end; str target unit should give OpenMM unit, while
            # Pint target unit should give Pint unit
            swap = not isinstance(unit_, str)

            # Convert str or Pint target unit (unit_) to OpenMM unit
            # (unit__)
            unit_ = ureg.Unit(unit_)
            try:
                unit__ = unit.dimensionless
                for u, p in unit_._units.items():
                    unit__ *= getattr(unit, u) ** p
            except AttributeError:
                emsg = ("strip_unit() relies on the pint module for "
                        "parsing units. At least one unit in 'unit' is "
                        "not defined the same way in openmm.unit and pint, "
                        "so the unit conversion cannot be performed. Try "
                        "specifying a openmm.unit.Quantity object for "
                        "'unit' instead.")
                raise ValueError(emsg)
        value = value.value_in_unit(unit__)
        if swap:
            unit_, unit__ = unit__, unit_

    else:
        unit__ = unit_

    return value, unit__


"""
OpenMM physical constants and unit conversions
==============================================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains physical constants and functions for unit
reduction.
"""

from openmm import unit

from ..algorithm import unit as u

VACUUM_PERMITTIVITY = 8.854187812813e-12 * unit.farad / unit.meter

def get_scaling_factors(
        bases: dict[str, unit.Quantity], other: dict[str, list] = {}
    ) -> dict[str, unit.Quantity]:

    """
    Evaluates scaling factors for reduced units.

    .. seealso::

       This is an alias function. For more information, see
       :func:`mdcraft.algorithm.unit.get_scaling_factors`.
    """

    return u.get_scaling_factors(bases, other)

def get_lj_scaling_factors(
        bases: dict[str, unit.Quantity], other: dict[str, list] = {}
    ) -> dict[str, unit.Quantity]:

    """
    Evaluates scaling factors for reduced Lennard-Jones units.

    .. seealso::

       This is an alias function. For more information, see
       :func:`mdcraft.algorithm.unit.get_lj_scaling_factors`.
    """

    return u.get_lj_scaling_factors(bases, other)